# Copyright (c) 2020 Dell Inc. or its subsidiaries.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from copy import deepcopy
from unittest import mock

from cinder import exception
from cinder.tests.unit import test
from cinder.tests.unit.volume.drivers.dell_emc.powermax import (
    powermax_data as tpd)
from cinder.tests.unit.volume.drivers.dell_emc.powermax import (
    powermax_fake_objects as tpfo)
from cinder.volume.drivers.dell_emc.powermax import common
from cinder.volume.drivers.dell_emc.powermax import masking
from cinder.volume.drivers.dell_emc.powermax import provision
from cinder.volume.drivers.dell_emc.powermax import rest
from cinder.volume.drivers.dell_emc.powermax import utils
from cinder.volume import volume_utils


class PowerMaxMaskingTest(test.TestCase):
    def setUp(self):
        self.data = tpd.PowerMaxData()
        super(PowerMaxMaskingTest, self).setUp()
        self.mock_object(volume_utils, 'get_max_over_subscription_ratio')
        self.replication_device = self.data.sync_rep_device
        configuration = tpfo.FakeConfiguration(
            None, 'MaskingTests', 1, 1, san_ip='1.1.1.1',
            san_login='smc', powermax_array=self.data.array,
            powermax_srp='SRP_1', san_password='smc', san_api_port=8443,
            powermax_port_groups=[self.data.port_group_name_f],
            replication_device=self.replication_device)
        self._gather_info = common.PowerMaxCommon._gather_info
        self.mock_object(common.PowerMaxCommon, '_get_u4p_failover_info')
        self.mock_object(common.PowerMaxCommon, '_gather_info')
        self.mock_object(rest.PowerMaxRest, '_establish_rest_session',
                         return_value=tpfo.FakeRequestsSession())
        driver = common.PowerMaxCommon(
            'iSCSI', self.data.version, configuration=configuration)
        driver_fc = common.PowerMaxCommon(
            'FC', self.data.version, configuration=configuration)
        self.driver = driver
        self.driver_fc = driver_fc
        self.mask = self.driver.masking
        self.extra_specs = deepcopy(self.data.extra_specs)
        self.extra_specs[utils.PORTGROUPNAME] = self.data.port_group_name_i
        self.maskingviewdict = self.driver._populate_masking_dict(
            self.data.test_volume, self.data.connector, self.extra_specs)
        self.maskingviewdict['extra_specs'] = self.extra_specs
        self.maskingviewdict[utils.IS_MULTIATTACH] = False
        self.device_id = self.data.device_id
        self.volume_name = self.data.volume_details[0]['volume_identifier']

    def tearDown(self):
        super(PowerMaxMaskingTest, self).tearDown()
        common.PowerMaxCommon._gather_info = self._gather_info

    def test_sanity_port_group_check_none(self):
        self.assertRaises(
            exception.VolumeBackendAPIException,
            self.driver.masking._sanity_port_group_check,
            None, self.data.array)

    @mock.patch.object(rest.PowerMaxRest, 'get_portgroup', return_value=None)
    def test_sanity_port_group_check_invalid_portgroup(self, mock_pg):
        self.assertRaises(
            exception.VolumeBackendAPIException,
            self.driver.masking._sanity_port_group_check,
            None, self.data.array)

    @mock.patch.object(
        masking.PowerMaxMasking, '_check_director_and_port_status')
    @mock.patch.object(rest.PowerMaxRest, 'get_portgroup',
                       return_value=tpd.PowerMaxData.portgroup)
    def test_sanity_port_group_check(self, mock_pg, mock_check):
        self.driver.masking._sanity_port_group_check(
            self.data.port_group_name_f, self.data.array)

    @mock.patch.object(masking.PowerMaxMasking,
                       'get_or_create_masking_view_and_map_lun')
    def test_setup_masking_view(self, mock_get_or_create_mv):
        self.driver.masking.setup_masking_view(
            self.data.array, self.data.test_volume,
            self.maskingviewdict, self.extra_specs)
        mock_get_or_create_mv.assert_called_once()

    @mock.patch.object(masking.PowerMaxMasking, '_validate_attach',
                       return_value=True)
    @mock.patch.object(masking.PowerMaxMasking,
                       '_check_adding_volume_to_storage_group')
    @mock.patch.object(masking.PowerMaxMasking, '_move_vol_from_default_sg',
                       return_value=None)
    @mock.patch.object(masking.PowerMaxMasking, '_get_or_create_masking_view',
                       side_effect=[None, 'Error in masking view retrieval',
                                    exception.VolumeBackendAPIException])
    @mock.patch.object(rest.PowerMaxRest, 'get_element_from_masking_view',
                       side_effect=[tpd.PowerMaxData.port_group_name_i,
                                    Exception('Exception')])
    def test_get_or_create_masking_view_and_map_lun(
            self, mock_masking_view_element, mock_masking, mock_move,
            mock_add_volume, mock_validate):
        rollback_dict = (
            self.driver.masking.get_or_create_masking_view_and_map_lun(
                self.data.array, self.data.test_volume,
                self.maskingviewdict['maskingview_name'],
                self.maskingviewdict, self.extra_specs))
        self.assertEqual(self.maskingviewdict, rollback_dict)
        self.assertRaises(
            exception.VolumeBackendAPIException,
            self.driver.masking.get_or_create_masking_view_and_map_lun,
            self.data.array, self.data.test_volume,
            self.maskingviewdict['maskingview_name'],
            self.maskingviewdict, self.extra_specs)
        self.maskingviewdict['slo'] = None
        self.assertRaises(
            exception.VolumeBackendAPIException,
            self.driver.masking.get_or_create_masking_view_and_map_lun,
            self.data.array, self.data.test_volume,
            self.maskingviewdict['maskingview_name'],
            self.maskingviewdict, self.extra_specs)

    @mock.patch.object(masking.PowerMaxMasking,
                       '_check_adding_volume_to_storage_group',
                       return_value=None)
    @mock.patch.object(
        rest.PowerMaxRest, 'move_volume_between_storage_groups',
        side_effect=[None, exception.VolumeBackendAPIException(data='')])
    @mock.patch.object(rest.PowerMaxRest, 'is_volume_in_storagegroup',
                       side_effect=[True, False, True])
    def test_move_vol_from_default_sg(
            self, mock_volume_in_sg, mock_move_volume, mock_add):
        msg = None
        for x in range(0, 2):
            msg = self.driver.masking._move_vol_from_default_sg(
                self.data.array, self.device_id, self.volume_name,
                self.data.defaultstoragegroup_name,
                self.data.storagegroup_name_i, self.extra_specs)
        mock_move_volume.assert_called_once()
        mock_add.assert_called_once()
        self.assertIsNone(msg)
        msg = self.driver.masking._move_vol_from_default_sg(
            self.data.array, self.device_id, self.volume_name,
            self.data.defaultstoragegroup_name,
            self.data.storagegroup_name_i, self.extra_specs)
        self.assertIsNotNone(msg)

    @mock.patch.object(rest.PowerMaxRest, 'modify_storage_group',
                       return_value=(200, tpfo.tpd.PowerMaxData.job_list[0]))
    @mock.patch.object(rest.PowerMaxRest, 'remove_child_sg_from_parent_sg')
    @mock.patch.object(masking.PowerMaxMasking, 'get_parent_sg_from_child',
                       side_effect=[None, tpd.PowerMaxData.parent_sg_f])
    @mock.patch.object(
        rest.PowerMaxRest, 'get_num_vols_in_sg', side_effect=[2, 1, 1])
    def test_move_volume_between_storage_groups(
            self, mock_num, mock_parent, mock_rm, mck_mod):
        for x in range(0, 3):
            self.driver.masking.move_volume_between_storage_groups(
                self.data.array, self.data.device_id,
                self.data.storagegroup_name_i, self.data.storagegroup_name_f,
                self.data.extra_specs)
        mock_rm.assert_called_once()
        ref_payload = (
            {"executionOption": "ASYNCHRONOUS",
             "editStorageGroupActionParam": {
                 "moveVolumeToStorageGroupParam": {
                     "volumeId": [self.data.device_id],
                     "storageGroupId": self.data.storagegroup_name_f,
                     "force": 'false'}}})
        mck_mod.assert_called_with(
            self.data.array, self.data.storagegroup_name_i, ref_payload)

    @mock.patch.object(rest.PowerMaxRest, 'move_volume_between_storage_groups')
    def test_move_volume_between_storage_groups_same_target(self, mock_mv):
        src = 'OS-SG'
        tgt = 'OS-SG'
        self.driver.masking.move_volume_between_storage_groups(
            self.data.array, self.data.device_id,
            src, tgt, self.data.extra_specs)
        mock_mv.assert_not_called()

    @mock.patch.object(rest.PowerMaxRest, 'remove_child_sg_from_parent_sg')
    @mock.patch.object(masking.PowerMaxMasking, 'get_parent_sg_from_child',
                       side_effect=[None, tpd.PowerMaxData.parent_sg_f])
    @mock.patch.object(rest.PowerMaxRest, 'move_volume_between_storage_groups')
    @mock.patch.object(
        rest.PowerMaxRest, 'get_num_vols_in_sg', return_value=1)
    def test_force_move_volume_between_storage_groups(
            self, mock_num, mock_move, mock_parent, mock_rm):

        self.driver.masking.move_volume_between_storage_groups(
            self.data.array, self.data.device_id,
            self.data.storagegroup_name_i, self.data.storagegroup_name_f,
            self.data.extra_specs, force=True)

        mock_move.assert_called_once_with(
            self.data.array, self.data.device_id,
            self.data.storagegroup_name_i, self.data.storagegroup_name_f,
            self.data.extra_specs, True)

    @mock.patch.object(
        masking.PowerMaxMasking, '_check_director_and_port_status')
    @mock.patch.object(rest.PowerMaxRest, 'get_masking_view',
                       side_effect=[tpd.PowerMaxData.maskingview,
                                    tpd.PowerMaxData.maskingview, None])
    @mock.patch.object(
        masking.PowerMaxMasking, '_validate_existing_masking_view',
        side_effect=[(tpd.PowerMaxData.maskingview[1]['storageGroupId'],
                      None), (None, 'Error Message')])
    @mock.patch.object(masking.PowerMaxMasking, '_create_new_masking_view',
                       return_value=None)
    def test_get_or_create_masking_view(self, mock_create_mv, mock_validate_mv,
                                        mock_get_mv, mock_check):
        for x in range(0, 3):
            self.driver.masking._get_or_create_masking_view(
                self.data.array, self.maskingviewdict,
                self.data.defaultstoragegroup_name, self.extra_specs)
        mock_create_mv.assert_called_once()

    @mock.patch.object(
        masking.PowerMaxMasking, '_get_or_create_initiator_group',
        side_effect=[(None, 'Initiator group error'), (None, None),
                     (None, None), (None, None), (None, None),
                     (None, None), (None, None), (None, None)])
    @mock.patch.object(
        masking.PowerMaxMasking, '_get_or_create_storage_group',
        side_effect=['Storage group not found', None,
                     'Storage group not found', 'Storage group not found',
                     None, None, None, None, None, None, None])
    @mock.patch.object(
        masking.PowerMaxMasking, '_move_vol_from_default_sg',
        side_effect=['Storage group error', None, 'Storage group error',
                     None])
    @mock.patch.object(
        masking.PowerMaxMasking, 'create_masking_view', return_value=None)
    def test_create_new_masking_view(
            self, mock_create_mv, mock_move, mock_create_SG, mock_create_IG):
        for x in range(0, 6):
            self.driver.masking._create_new_masking_view(
                self.data.array, self.maskingviewdict,
                self.maskingviewdict['maskingview_name'],
                self.data.defaultstoragegroup_name, self.extra_specs)
        mock_create_mv.assert_called_once()

    @mock.patch.object(
        masking.PowerMaxMasking, '_check_existing_storage_group',
        side_effect=[(tpd.PowerMaxData.storagegroup_name_i, None),
                     (tpd.PowerMaxData.storagegroup_name_i, None),
                     (None, 'Error Checking existing storage group')])
    @mock.patch.object(
        rest.PowerMaxRest, 'get_element_from_masking_view',
        return_value=tpd.PowerMaxData.port_group_name_i)
    @mock.patch.object(
        masking.PowerMaxMasking, '_check_port_group',
        side_effect=[(None, None), (None, 'Error checking pg')])
    @mock.patch.object(
        masking.PowerMaxMasking, '_check_existing_initiator_group',
        return_value=(tpd.PowerMaxData.initiatorgroup_name_i, None))
    def test_validate_existing_masking_view(
            self, mock_check_ig, mock_check_pg, mock_get_mv_element,
            mock_check_sg):
        for x in range(0, 3):
            self.driver.masking._validate_existing_masking_view(
                self.data.array, self.maskingviewdict,
                self.maskingviewdict['maskingview_name'],
                self.data.defaultstoragegroup_name, self.extra_specs)
        self.assertEqual(3, mock_check_sg.call_count)
        mock_get_mv_element.assert_called_with(
            self.data.array, self.maskingviewdict['maskingview_name'],
            portgroup=True)
        mock_check_ig.assert_called_once()

    @mock.patch.object(
        rest.PowerMaxRest, 'get_storage_group',
        side_effect=[tpd.PowerMaxData.storagegroup_name_i, None,
                     tpd.PowerMaxData.storagegroup_name_i])
    @mock.patch.object(
        provision.PowerMaxProvision, 'create_storage_group',
        side_effect=[tpd.PowerMaxData.storagegroup_name_i, None])
    def test_get_or_create_storage_group(self, mock_sg, mock_get_sg):
        for x in range(0, 2):
            self.driver.masking._get_or_create_storage_group(
                self.data.array, self.maskingviewdict,
                self.data.storagegroup_name_i, self.extra_specs)
        self.assertEqual(3, mock_get_sg.call_count)
        self.assertEqual(1, mock_sg.call_count)

    @mock.patch.object(
        rest.PowerMaxRest, 'get_storage_group',
        side_effect=[None, tpd.PowerMaxData.storagegroup_name_i])
    @mock.patch.object(
        provision.PowerMaxProvision, 'create_storage_group',
        side_effect=[tpd.PowerMaxData.storagegroup_name_i])
    def test_get_or_create_storage_group_is_parent(self, mock_sg, mock_get_sg):
        self.driver.masking._get_or_create_storage_group(
            self.data.array, self.maskingviewdict,
            self.data.storagegroup_name_i, self.extra_specs, True)
        self.assertEqual(2, mock_get_sg.call_count)
        self.assertEqual(1, mock_sg.call_count)

    @mock.patch.object(
        rest.PowerMaxRest, 'update_storagegroup_qos')
    @mock.patch.object(
        rest.PowerMaxRest, 'get_storage_group',
        return_value=tpd.PowerMaxData.storagegroup_name_i)
    def test_get_or_create_storage_group_is_parent_qos(
            self, mock_sg, mock_update_sg):
        self.driver.masking._get_or_create_storage_group(
            self.data.array, self.maskingviewdict,
            self.data.storagegroup_name_i, self.data.extra_specs_qos, True)
        mock_update_sg.assert_not_called()

    @mock.patch.object(
        rest.PowerMaxRest, 'update_storagegroup_qos')
    @mock.patch.object(
        rest.PowerMaxRest, 'get_storage_group',
        return_value=tpd.PowerMaxData.storagegroup_name_i)
    def test_get_or_create_storage_group_is_child_qos(
            self, mock_sg, mock_update_sg):
        self.driver.masking._get_or_create_storage_group(
            self.data.array, self.maskingviewdict,
            self.data.storagegroup_name_i, self.data.extra_specs_qos, False)
        mock_update_sg.assert_called_once()

    @mock.patch.object(masking.PowerMaxMasking, '_move_vol_from_default_sg',
                       return_value=None)
    @mock.patch.object(masking.PowerMaxMasking, '_get_or_create_storage_group',
                       return_value=None)
    @mock.patch.object(rest.PowerMaxRest, 'get_element_from_masking_view',
                       return_value=tpd.PowerMaxData.parent_sg_i)
    @mock.patch.object(rest.PowerMaxRest, 'is_child_sg_in_parent_sg',
                       side_effect=[True, False])
    @mock.patch.object(masking.PowerMaxMasking,
                       '_check_add_child_sg_to_parent_sg', return_value=None)
    def test_check_existing_storage_group_success(
            self, mock_add_sg, mock_is_child, mock_get_mv_element,
            mock_create_sg, mock_move):
        masking_view_dict = deepcopy(self.data.masking_view_dict)
        masking_view_dict['extra_specs'] = self.data.extra_specs

        with mock.patch.object(
                self.driver.rest, 'get_storage_group',
                side_effect=[tpd.PowerMaxData.parent_sg_i,
                             tpd.PowerMaxData.storagegroup_name_i]):
            _, msg = (self.driver.masking._check_existing_storage_group(
                self.data.array, self.maskingviewdict['maskingview_name'],
                self.data.defaultstoragegroup_name, masking_view_dict,
                self.data.extra_specs))
            self.assertIsNone(msg)
            mock_create_sg.assert_not_called()

        with mock.patch.object(self.driver.rest, 'get_storage_group',
                               side_effect=[
                                   tpd.PowerMaxData.parent_sg_i, None]):
            _, msg = (self.driver.masking._check_existing_storage_group(
                self.data.array, self.maskingviewdict['maskingview_name'],
                self.data.defaultstoragegroup_name, masking_view_dict,
                self.data.extra_specs))
            self.assertIsNone(msg)
            mock_create_sg.assert_called_once_with(
                self.data.array, masking_view_dict,
                tpd.PowerMaxData.storagegroup_name_f,
                self.data.extra_specs)

    @mock.patch.object(masking.PowerMaxMasking, '_move_vol_from_default_sg',
                       side_effect=[None, 'Error Message'])
    @mock.patch.object(rest.PowerMaxRest, 'is_child_sg_in_parent_sg',
                       side_effect=[True, False, False])
    @mock.patch.object(rest.PowerMaxRest, 'get_element_from_masking_view',
                       return_value=tpd.PowerMaxData.parent_sg_i)
    @mock.patch.object(rest.PowerMaxRest, 'get_storage_group',
                       side_effect=[
                           None, tpd.PowerMaxData.parent_sg_i, None,
                           tpd.PowerMaxData.parent_sg_i, None,
                           tpd.PowerMaxData.parent_sg_i, None])
    def test_check_existing_storage_group_failed(
            self, mock_get_sg, mock_get_mv_element, mock_child, mock_move):
        masking_view_dict = deepcopy(self.data.masking_view_dict)
        masking_view_dict['extra_specs'] = self.data.extra_specs
        for x in range(0, 4):
            _, msg = (self.driver.masking._check_existing_storage_group(
                self.data.array, self.maskingviewdict['maskingview_name'],
                self.data.defaultstoragegroup_name, masking_view_dict,
                self.data.extra_specs))
            self.assertIsNotNone(msg)
        self.assertEqual(7, mock_get_sg.call_count)
        self.assertEqual(1, mock_move.call_count)

    @mock.patch.object(
        masking.PowerMaxMasking, '_check_director_and_port_status')
    @mock.patch.object(
        rest.PowerMaxRest, 'get_portgroup',
        side_effect=([tpd.PowerMaxData.port_group_name_i, None]))
    def test_check_port_group(
            self, mock_get_pg, mock_check):
        for x in range(0, 2):
            _, msg = self.driver.masking._check_port_group(
                self.data.array, self.maskingviewdict['maskingview_name'])
        self.assertIsNotNone(msg)
        self.assertEqual(2, mock_get_pg.call_count)

    @mock.patch.object(
        rest.PowerMaxRest, 'get_initiator_group',
        side_effect=[None, tpd.PowerMaxData.initiator_group_iscsi])
    @mock.patch.object(
        masking.PowerMaxMasking, '_find_initiator_group',
        side_effect=[tpd.PowerMaxData.initiatorgroup_name_i, None, None])
    @mock.patch.object(
        masking.PowerMaxMasking, '_create_initiator_group',
        side_effect=([tpd.PowerMaxData.initiatorgroup_name_i, None]))
    def test_get_or_create_initiator_group(
            self, mock_create_ig, mock_find_ig, mock_get_ig):
        self.driver.masking._get_or_create_initiator_group(
            self.data.array, self.data.initiatorgroup_name_i,
            self.data.connector, self.extra_specs)
        mock_create_ig.assert_not_called()
        found_init_group_name, msg = (
            self.driver.masking._get_or_create_initiator_group(
                self.data.array, self.data.initiatorgroup_name_i,
                self.data.connector, self.extra_specs))
        self.assertIsNone(msg)
        found_init_group_name, msg = (
            self.driver.masking._get_or_create_initiator_group(
                self.data.array, self.data.initiatorgroup_name_i,
                self.data.connector, self.extra_specs))
        self.assertIsNone(msg)

    @mock.patch.object(
        rest.PowerMaxRest, 'get_initiator_group',
        return_value=tpd.PowerMaxData.initiator_group_iscsi)
    @mock.patch.object(
        masking.PowerMaxMasking, '_find_initiator_group',
        return_value=None)
    @mock.patch.object(
        masking.PowerMaxMasking, '_create_initiator_group')
    def test_get_or_create_initiator_group_not_logged_in(
            self, mock_create_ig, mock_find_ig, mock_get_ig):
        found_init_group, msg = (
            self.driver.masking._get_or_create_initiator_group(
                self.data.array, self.data.initiatorgroup_name_i,
                self.data.connector, self.extra_specs))
        mock_create_ig.assert_not_called()
        self.assertIsNone(msg)

    @mock.patch.object(
        rest.PowerMaxRest, 'get_initiator_group',
        side_effect=[tpd.PowerMaxData.initiator_group_fc,
                     tpd.PowerMaxData.initiator_group_empty])
    @mock.patch.object(
        masking.PowerMaxMasking, '_find_initiator_group',
        return_value=None)
    @mock.patch.object(
        masking.PowerMaxMasking, '_create_initiator_group')
    def test_get_or_create_initiator_group_not_logged_in_errors(
            self, mock_create_ig, mock_find_ig, mock_get_ig):
        found_init_group, msg = (
            self.driver.masking._get_or_create_initiator_group(
                self.data.array, self.data.initiatorgroup_name_i,
                self.data.connector, self.extra_specs))
        expected_msg = (
            "Found initiator group OS-HostX-I-IG, but could not find "
            "initiator_names ['iqn.1993-08.org.debian:01:222'] in "
            "the login table. The contained initiators ['123456789012345'] "
            "do match up with those in the connector object. Delete initiator "
            "group OS-HostX-I-IG and retry.")
        self.assertEqual(expected_msg, msg)
        mock_create_ig.assert_not_called()
        found_init_group, msg = (
            self.driver.masking._get_or_create_initiator_group(
                self.data.array, self.data.initiatorgroup_name_i,
                self.data.connector, self.extra_specs))
        expected_msg = (
            "Found initiator group OS-HostX-I-IG, but could not find "
            "initiator_names ['iqn.1993-08.org.debian:01:222'] in the login "
            "table. There are no initiators in OS-HostX-I-IG. Delete "
            "initiator group OS-HostX-I-IG and retry.")
        self.assertEqual(expected_msg, msg)
        mock_create_ig.assert_not_called()

    def test_check_existing_initiator_group(self):
        with mock.patch.object(
                rest.PowerMaxRest, 'get_element_from_masking_view',
                return_value=tpd.PowerMaxData.initiatorgroup_name_f):
            ig_from_mv, msg = (
                self.driver.masking._check_existing_initiator_group(
                    self.data.array, self.maskingviewdict['maskingview_name'],
                    self.maskingviewdict, self.data.storagegroup_name_i,
                    self.data.port_group_name_i, self.extra_specs))
            self.assertEqual(self.data.initiatorgroup_name_f, ig_from_mv)

    def test_check_adding_volume_to_storage_group(self):
        with mock.patch.object(
                masking.PowerMaxMasking, '_create_initiator_group'):
            with mock.patch.object(
                rest.PowerMaxRest, 'is_volume_in_storagegroup',
                    side_effect=[True, False]):
                msg = (
                    self.driver.masking._check_adding_volume_to_storage_group(
                        self.data.array, self.device_id,
                        self.data.storagegroup_name_i,
                        self.maskingviewdict[utils.VOL_NAME],
                        self.maskingviewdict[utils.EXTRA_SPECS]))
                self.assertIsNone(msg)
                msg = (
                    self.driver.masking._check_adding_volume_to_storage_group(
                        self.data.array, self.device_id,
                        self.data.storagegroup_name_i,
                        self.maskingviewdict[utils.VOL_NAME],
                        self.maskingviewdict[utils.EXTRA_SPECS]))

    @mock.patch.object(rest.PowerMaxRest, 'add_vol_to_sg')
    def test_add_volume_to_storage_group(self, mock_add_volume):
        self.driver.masking.add_volume_to_storage_group(
            self.data.array, self.device_id, self.data.storagegroup_name_i,
            self.volume_name, self.extra_specs)
        mock_add_volume.assert_called_once()

    @mock.patch.object(rest.PowerMaxRest, 'remove_vol_from_sg')
    def test_remove_vol_from_storage_group(self, mock_remove_volume):
        with mock.patch.object(
                rest.PowerMaxRest, 'is_volume_in_storagegroup',
                side_effect=[False, True]):
            self.driver.masking.remove_vol_from_storage_group(
                self.data.array, self.device_id, self.data.storagegroup_name_i,
                self.volume_name, self.extra_specs)
            mock_remove_volume.assert_called_once()
            self.assertRaises(
                exception.VolumeBackendAPIException,
                self.driver.masking.remove_vol_from_storage_group,
                self.data.array, self.device_id, self.data.storagegroup_name_i,
                self.volume_name, self.extra_specs)

    def test_find_initiator_names(self):
        foundinitiatornames = self.driver.masking.find_initiator_names(
            self.data.connector)
        self.assertEqual(self.data.connector['initiator'],
                         foundinitiatornames[0])
        foundinitiatornames = self.driver_fc.masking.find_initiator_names(
            self.data.connector)
        self.assertEqual(self.data.connector['wwpns'][0],
                         foundinitiatornames[0])
        connector = {'ip': self.data.ip, 'initiator': None, 'host': 'HostX'}
        self.assertRaises(
            exception.VolumeBackendAPIException,
            self.driver.masking.find_initiator_names, connector)
        self.assertRaises(
            exception.VolumeBackendAPIException,
            self.driver_fc.masking.find_initiator_names, connector)

    def test_find_initiator_group_found(self):
        with mock.patch.object(
                rest.PowerMaxRest, 'get_initiator_list',
                return_value=self.data.initiator_list[2]['initiatorId']):
            with mock.patch.object(
                    rest.PowerMaxRest, 'get_initiator_group_from_initiator',
                    return_value=self.data.initiator_list):
                found_init_group_nam = (
                    self.driver.masking._find_initiator_group(
                        self.data.array, ['FA-1D:4:123456789012345']))
                self.assertEqual(self.data.initiator_list,
                                 found_init_group_nam)

    def test_find_initiator_group_not_found(self):
        with mock.patch.object(
                rest.PowerMaxRest, 'get_initiator_list',
                return_value=self.data.initiator_list[2]['initiatorId']):
            with mock.patch.object(
                    rest.PowerMaxRest, 'get_initiator_group_from_initiator',
                    return_value=None):
                found_init_group_nam = (
                    self.driver.masking._find_initiator_group(
                        self.data.array, ['Error']))
                self.assertIsNone(found_init_group_nam)

    def test_create_masking_view(self):
        with mock.patch.object(rest.PowerMaxRest, 'create_masking_view',
                               side_effect=[None, Exception]):
            error_message = self.driver.masking.create_masking_view(
                self.data.array, self.maskingviewdict['maskingview_name'],
                self.data.storagegroup_name_i, self.data.port_group_name_i,
                self.data.initiatorgroup_name_i, self.extra_specs)
            self.assertIsNone(error_message)
            error_message = self.driver.masking.create_masking_view(
                self.data.array, self.maskingviewdict['maskingview_name'],
                self.data.storagegroup_name_i, self.data.port_group_name_i,
                self.data.initiatorgroup_name_i, self.extra_specs)
            self.assertIsNotNone(error_message)

    @mock.patch.object(masking.PowerMaxMasking,
                       '_return_volume_to_fast_managed_group')
    @mock.patch.object(masking.PowerMaxMasking, '_check_ig_rollback')
    def test_check_if_rollback_action_for_masking_required(
            self, mock_check_ig, mock_return):
        with mock.patch.object(rest.PowerMaxRest,
                               'get_storage_groups_from_volume',
                               side_effect=[
                                   exception.VolumeBackendAPIException,
                                   self.data.storagegroup_list,
                                   self.data.storagegroup_list, None,
                                   None, ]):
            self.assertRaises(
                exception.VolumeBackendAPIException,
                self.mask.check_if_rollback_action_for_masking_required,
                self.data.array, self.data.test_volume,
                self.device_id, self.maskingviewdict)
            with mock.patch.object(masking.PowerMaxMasking,
                                   'remove_and_reset_members'):
                self.maskingviewdict[
                    'default_sg_name'] = self.data.defaultstoragegroup_name
                self.mask.check_if_rollback_action_for_masking_required(
                    self.data.array, self.data.test_volume,
                    self.device_id, self.maskingviewdict)
                # Multiattach case
                self.mask.check_if_rollback_action_for_masking_required(
                    self.data.array, self.data.test_volume,
                    self.device_id, self.data.masking_view_dict_multiattach)
                mock_return.assert_called_once()

    @mock.patch.object(masking.PowerMaxMasking, '_recreate_masking_view')
    @mock.patch.object(rest.PowerMaxRest, 'get_initiator_group',
                       return_value=True)
    def test_verify_initiator_group_from_masking_view(
            self, mock_get_ig, mock_recreate_mv):
        mv_dict = deepcopy(self.maskingviewdict)
        mv_dict['initiator_check'] = True
        self.mask._verify_initiator_group_from_masking_view(
            self.data.array, mv_dict['maskingview_name'],
            mv_dict, self.data.initiatorgroup_name_i,
            self.data.storagegroup_name_i, self.data.port_group_name_i,
            self.extra_specs)
        mock_recreate_mv.assert_called()

    @mock.patch.object(masking.PowerMaxMasking, '_recreate_masking_view')
    @mock.patch.object(rest.PowerMaxRest, 'get_initiator_group',
                       return_value=True)
    @mock.patch.object(
        masking.PowerMaxMasking, '_find_initiator_group',
        return_value=tpd.PowerMaxData.initiatorgroup_name_i)
    def test_verify_initiator_group_from_masking_view_no_recreate(
            self, mock_find_ig, mock_get_ig, mock_recreate):
        mv_dict = deepcopy(self.maskingviewdict)
        mv_dict['initiator_check'] = False
        self.assertRaises(
            exception.VolumeBackendAPIException,
            self.mask._verify_initiator_group_from_masking_view,
            self.data.array, mv_dict['maskingview_name'],
            mv_dict, 'OS-Wrong-Host-I-IG',
            self.data.storagegroup_name_i, self.data.port_group_name_i,
            self.extra_specs)
        mock_recreate.assert_not_called()

    @mock.patch.object(rest.PowerMaxRest, 'delete_initiator_group')
    @mock.patch.object(rest.PowerMaxRest, 'get_initiator_group',
                       return_value=True)
    def test_recreate_masking_view(
            self, mock_get_ig, mock_delete_ig):

        ig_from_conn = self.data.initiatorgroup_name_i
        ig_from_mv = self.data.initiatorgroup_name_i
        ig_openstack = self.data.initiatorgroup_name_i

        self.mask._recreate_masking_view(
            self.data.array, ig_from_conn, ig_from_mv,
            ig_openstack, self.data.masking_view_name_i, [self.data.initiator],
            self.data.storagegroup_name_i, self.data.port_group_name_i,
            self.extra_specs)
        mock_delete_ig.assert_not_called()

    @mock.patch.object(rest.PowerMaxRest, 'delete_initiator_group')
    @mock.patch.object(rest.PowerMaxRest, 'get_initiator_group',
                       return_value=True)
    def test_recreate_masking_view_no_ig_from_connector(
            self, mock_get_ig, mock_delete_ig):

        ig_from_mv = self.data.initiatorgroup_name_i
        ig_openstack = self.data.initiatorgroup_name_i

        self.mask._recreate_masking_view(
            self.data.array, None, ig_from_mv,
            ig_openstack, self.data.masking_view_name_i, [self.data.initiator],
            self.data.storagegroup_name_i, self.data.port_group_name_i,
            self.extra_specs)
        mock_delete_ig.assert_called()

    @mock.patch.object(rest.PowerMaxRest, 'create_masking_view')
    @mock.patch.object(rest.PowerMaxRest, 'get_initiator_group',
                       return_value=True)
    def test_recreate_masking_view_wrong_host(
            self, mock_get_ig, mock_create_mv):

        ig_from_conn = 'OS-Wrong-Host-I-IG'
        ig_from_mv = self.data.initiatorgroup_name_i
        ig_openstack = self.data.initiatorgroup_name_i

        self.mask._recreate_masking_view(
            self.data.array, ig_from_conn, ig_from_mv,
            ig_openstack, self.data.masking_view_name_i, [self.data.initiator],
            self.data.storagegroup_name_i, self.data.port_group_name_i,
            self.extra_specs)
        mock_create_mv.assert_called()

    @mock.patch.object(rest.PowerMaxRest, 'delete_masking_view')
    @mock.patch.object(rest.PowerMaxRest, 'delete_initiator_group')
    @mock.patch.object(rest.PowerMaxRest, 'get_initiator_group',
                       return_value=True)
    @mock.patch.object(
        masking.PowerMaxMasking, '_find_initiator_group',
        return_value=tpd.PowerMaxData.initiatorgroup_name_i)
    def test_recreate_masking_view_delete_mv(
            self, mock_find_ig, mock_get_ig, mock_delete_ig, mock_delete_mv):

        mock_delete_mv.side_effect = [None, Exception]
        mv_dict = deepcopy(self.maskingviewdict)
        mv_dict['initiator_check'] = True
        verify_flag = self.mask._verify_initiator_group_from_masking_view(
            self.data.array, mv_dict['maskingview_name'],
            mv_dict, 'OS-Wrong-Host-I-IG',
            self.data.storagegroup_name_i, self.data.port_group_name_i,
            self.extra_specs)
        mock_delete_mv.assert_called()
        self.assertTrue(verify_flag)

    @mock.patch.object(rest.PowerMaxRest, 'create_initiator_group')
    def test_create_initiator_group(self, mock_create_ig):
        initiator_names = self.mask.find_initiator_names(self.data.connector)
        ret_init_group_name = self.mask._create_initiator_group(
            self.data.array, self.data.initiatorgroup_name_i, initiator_names,
            self.extra_specs)
        self.assertEqual(self.data.initiatorgroup_name_i, ret_init_group_name)

    @mock.patch.object(rest.PowerMaxRest, 'create_initiator_group',
                       side_effect=([exception.VolumeBackendAPIException(
                           masking.CREATE_IG_ERROR)]))
    def test_create_initiator_group_exception(self, mock_create_ig):
        initiator_names = self.mask.find_initiator_names(self.data.connector)
        self.assertRaises(
            exception.VolumeBackendAPIException,
            self.mask._create_initiator_group,
            self.data.array, self.data.initiatorgroup_name_i, initiator_names,
            self.extra_specs)

    @mock.patch.object(masking.PowerMaxMasking,
                       '_last_volume_delete_initiator_group')
    def test_check_ig_rollback(self, mock_last_volume):
        with mock.patch.object(
                masking.PowerMaxMasking, '_find_initiator_group',
                side_effect=[None, 'FAKE-I-IG',
                             self.data.initiatorgroup_name_i]):
            for x in range(0, 2):
                self.mask._check_ig_rollback(self.data.array,
                                             self.data.initiatorgroup_name_i,
                                             self.data.connector)
            mock_last_volume.assert_not_called()
            self.mask._check_ig_rollback(
                self.data.array, self.data.initiatorgroup_name_i,
                self.data.connector)
            mock_last_volume.assert_called()

    @mock.patch.object(masking.PowerMaxMasking, '_cleanup_deletion')
    def test_remove_and_reset_members(self, mock_cleanup):
        self.mask.remove_and_reset_members(
            self.data.array, self.device_id, self.data.test_volume,
            self.volume_name, self.extra_specs, reset=False)
        mock_cleanup.assert_called_once()

    @mock.patch.object(
        rest.PowerMaxRest, 'get_storage_groups_from_volume',
        side_effect=[[tpd.PowerMaxData.storagegroup_name_i],
                     [tpd.PowerMaxData.storagegroup_name_i],
                     [tpd.PowerMaxData.storagegroup_name_i,
                      tpd.PowerMaxData.storagegroup_name_f]])
    @mock.patch.object(masking.PowerMaxMasking, 'remove_volume_from_sg')
    @mock.patch.object(masking.PowerMaxMasking,
                       'add_volume_to_default_storage_group')
    def test_cleanup_deletion(self, mock_add, mock_remove_vol, mock_get_sg):
        self.mask._cleanup_deletion(
            self.data.array, self.data.test_volume, self.device_id,
            self.volume_name, self.extra_specs, None, True, None)
        mock_add.assert_not_called()
        self.mask._cleanup_deletion(
            self.data.array, self.data.test_volume, self.device_id,
            self.volume_name, self.extra_specs,
            self.data.connector, True, None)
        mock_add.assert_not_called()
        self.mask._cleanup_deletion(
            self.data.array, self.data.test_volume, self.device_id,
            self.volume_name, self.extra_specs, None, True, None)
        mock_add.assert_called_once_with(
            self.data.array, self.device_id,
            self.volume_name, self.extra_specs, volume=self.data.test_volume)

    @mock.patch.object(masking.PowerMaxMasking, '_last_vol_in_sg')
    @mock.patch.object(masking.PowerMaxMasking, '_multiple_vols_in_sg')
    def test_remove_volume_from_sg(self, mock_multiple_vols, mock_last_vol):
        with mock.patch.object(
                rest.PowerMaxRest, 'get_masking_views_from_storage_group',
                return_value=None):
            with mock.patch.object(
                    rest.PowerMaxRest, 'get_num_vols_in_sg',
                    side_effect=[2, 1]):
                self.mask.remove_volume_from_sg(
                    self.data.array, self.device_id, self.volume_name,
                    self.data.defaultstoragegroup_name, self.extra_specs)
                mock_last_vol.assert_not_called()
                self.mask.remove_volume_from_sg(
                    self.data.array, self.device_id, self.volume_name,
                    self.data.defaultstoragegroup_name, self.extra_specs)
                mock_last_vol.assert_called()

    @mock.patch.object(masking.PowerMaxMasking, '_last_vol_in_sg')
    @mock.patch.object(masking.PowerMaxMasking, '_multiple_vols_in_sg')
    def test_remove_volume_from_sg_2(self, mock_multiple_vols, mock_last_vol):
        with mock.patch.object(rest.PowerMaxRest, 'is_volume_in_storagegroup',
                               return_value=True):
            with mock.patch.object(
                    rest.PowerMaxRest, 'get_masking_views_from_storage_group',
                    return_value=[self.data.masking_view_name_i]):
                with mock.patch.object(
                    rest.PowerMaxRest, 'get_num_vols_in_sg',
                        side_effect=[2, 1]):
                    self.mask.remove_volume_from_sg(
                        self.data.array, self.device_id, self.volume_name,
                        self.data.storagegroup_name_i, self.extra_specs)
                    mock_last_vol.assert_not_called()
                    self.mask.remove_volume_from_sg(
                        self.data.array, self.device_id, self.volume_name,
                        self.data.storagegroup_name_i, self.extra_specs)
                    mock_last_vol.assert_called()

    @mock.patch.object(masking.PowerMaxMasking, '_last_vol_masking_views',
                       return_value=True)
    @mock.patch.object(masking.PowerMaxMasking, '_last_vol_no_masking_views',
                       return_value=True)
    def test_last_vol_in_sg(self, mock_no_mv, mock_mv):
        mv_list = [self.data.masking_view_name_i,
                   self.data.masking_view_name_f]
        with mock.patch.object(rest.PowerMaxRest,
                               'get_masking_views_from_storage_group',
                               side_effect=[mv_list, []]):
            for x in range(0, 2):
                self.mask._last_vol_in_sg(
                    self.data.array, self.device_id, self.volume_name,
                    self.data.storagegroup_name_i, self.extra_specs,
                    self.data.connector)
            self.assertEqual(1, mock_mv.call_count)
            self.assertEqual(1, mock_no_mv.call_count)

    @mock.patch.object(masking.PowerMaxMasking,
                       '_remove_last_vol_and_delete_sg')
    @mock.patch.object(masking.PowerMaxMasking,
                       '_delete_cascaded_storage_groups')
    @mock.patch.object(rest.PowerMaxRest, 'get_num_vols_in_sg',
                       side_effect=[1, 3])
    @mock.patch.object(rest.PowerMaxRest, 'delete_storage_group')
    @mock.patch.object(masking.PowerMaxMasking, 'get_parent_sg_from_child',
                       side_effect=[None, 'parent_sg_name', 'parent_sg_name'])
    def test_last_vol_no_masking_views(
            self, mock_get_parent, mock_delete, mock_num_vols,
            mock_delete_casc, mock_remove):
        for x in range(0, 3):
            self.mask._last_vol_no_masking_views(
                self.data.array, self.data.storagegroup_name_i,
                self.device_id, self.volume_name, self.extra_specs,
                False)
        self.assertEqual(1, mock_delete.call_count)
        self.assertEqual(1, mock_delete_casc.call_count)
        self.assertEqual(1, mock_remove.call_count)

    @mock.patch.object(masking.PowerMaxMasking,
                       '_remove_last_vol_and_delete_sg')
    @mock.patch.object(masking.PowerMaxMasking, '_delete_mv_ig_and_sg')
    @mock.patch.object(masking.PowerMaxMasking, '_get_num_vols_from_mv',
                       side_effect=[(1, 'parent_name'), (3, 'parent_name')])
    def test_last_vol_masking_views(
            self, mock_num_vols, mock_delete_all, mock_remove):
        for x in range(0, 2):
            self.mask._last_vol_masking_views(
                self.data.array, self.data.storagegroup_name_i,
                [self.data.masking_view_name_i], self.device_id,
                self.volume_name, self.extra_specs, self.data.connector,
                True)
        self.assertEqual(1, mock_delete_all.call_count)
        self.assertEqual(1, mock_remove.call_count)

    @mock.patch.object(masking.PowerMaxMasking,
                       'add_volume_to_default_storage_group')
    @mock.patch.object(rest.PowerMaxRest, 'get_num_vols_in_sg')
    @mock.patch.object(masking.PowerMaxMasking,
                       'remove_vol_from_storage_group')
    def test_multiple_vols_in_sg(self, mock_remove_vol, mock_get_volumes,
                                 mock_add):
        self.mask._multiple_vols_in_sg(
            self.data.array, self.device_id, self.data.storagegroup_name_i,
            self.volume_name, self.extra_specs, False)
        mock_remove_vol.assert_called_once()
        self.mask._multiple_vols_in_sg(
            self.data.array, self.device_id, self.data.storagegroup_name_i,
            self.volume_name, self.extra_specs, True)
        mock_add.assert_called_once()

    @mock.patch.object(rest.PowerMaxRest, 'get_element_from_masking_view')
    @mock.patch.object(masking.PowerMaxMasking,
                       '_last_volume_delete_masking_view')
    @mock.patch.object(masking.PowerMaxMasking,
                       '_last_volume_delete_initiator_group')
    @mock.patch.object(masking.PowerMaxMasking,
                       '_delete_cascaded_storage_groups')
    def test_delete_mv_ig_and_sg(self, mock_delete_sg, mock_delete_ig,
                                 mock_delete_mv, mock_get_element):
        self.mask._delete_mv_ig_and_sg(
            self.data.array, self.data.device_id,
            self.data.masking_view_name_i,
            self.data.storagegroup_name_i, self.data.parent_sg_i,
            self.data.connector, True, self.data.extra_specs)
        mock_delete_sg.assert_called_once()

    @mock.patch.object(rest.PowerMaxRest, 'delete_masking_view')
    def test_last_volume_delete_masking_view(self, mock_delete_mv):
        self.mask._last_volume_delete_masking_view(
            self.data.array, self.data.masking_view_name_i)
        mock_delete_mv.assert_called_once()

    @mock.patch.object(
        rest.PowerMaxRest, 'get_storage_groups_from_volume',
        return_value=list())
    @mock.patch.object(masking.PowerMaxMasking,
                       'return_volume_to_volume_group')
    @mock.patch.object(rest.PowerMaxRest, 'move_volume_between_storage_groups')
    @mock.patch.object(masking.PowerMaxMasking,
                       'get_or_create_default_storage_group')
    @mock.patch.object(masking.PowerMaxMasking, 'add_volume_to_storage_group')
    def test_add_volume_to_default_storage_group(
            self, mock_add_sg, mock_get_sg, mock_move, mock_return, mock_sgs):
        self.mask.add_volume_to_default_storage_group(
            self.data.array, self.device_id, self.volume_name,
            self.extra_specs)
        mock_add_sg.assert_called_once()
        self.mask.add_volume_to_default_storage_group(
            self.data.array, self.device_id, self.volume_name,
            self.extra_specs, src_sg=self.data.storagegroup_name_i)
        mock_move.assert_called_once()
        vol_grp_member = deepcopy(self.data.test_volume)
        vol_grp_member.group_id = self.data.test_vol_grp_name_id_only
        self.mask.add_volume_to_default_storage_group(
            self.data.array, self.device_id, self.volume_name,
            self.extra_specs, volume=vol_grp_member)
        mock_return.assert_called_once()

    def test_add_volume_to_default_storage_group_next_gen(self):
        extra_specs = deepcopy(self.data.extra_specs)
        extra_specs.pop(utils.IS_RE, None)
        with mock.patch.object(rest.PowerMaxRest, 'is_next_gen_array',
                               return_value=True):
            with mock.patch.object(
                    self.mask,
                    'get_or_create_default_storage_group') as mock_get:
                self.mask.add_volume_to_default_storage_group(
                    self.data.array, self.device_id, self.volume_name,
                    extra_specs)
                mock_get.assert_called_once_with(
                    self.data.array, self.data.srp,
                    extra_specs[utils.SLO],
                    'NONE', extra_specs, False, False, None)

    @mock.patch.object(provision.PowerMaxProvision, 'create_storage_group')
    def test_get_or_create_default_storage_group(self, mock_create_sg):
        with mock.patch.object(
                rest.PowerMaxRest, 'get_vmax_default_storage_group',
                return_value=(None, self.data.storagegroup_name_i)):
            storage_group_name = self.mask.get_or_create_default_storage_group(
                self.data.array, self.data.srp, self.data.slo,
                self.data.workload, self.extra_specs)
            self.assertEqual(self.data.storagegroup_name_i, storage_group_name)
        with mock.patch.object(
                rest.PowerMaxRest, 'get_vmax_default_storage_group',
                return_value=('test_sg', self.data.storagegroup_name_i)):
            with mock.patch.object(
                rest.PowerMaxRest, 'get_masking_views_from_storage_group',
                    return_value=self.data.masking_view_name_i):
                self.assertRaises(
                    exception.VolumeBackendAPIException,
                    self.mask.get_or_create_default_storage_group,
                    self.data.array, self.data.srp, self.data.slo,
                    self.data.workload, self.extra_specs)

    @mock.patch.object(masking.PowerMaxMasking,
                       'add_volume_to_default_storage_group')
    @mock.patch.object(rest.PowerMaxRest, 'remove_child_sg_from_parent_sg')
    @mock.patch.object(rest.PowerMaxRest, 'delete_storage_group')
    @mock.patch.object(masking.PowerMaxMasking,
                       'remove_vol_from_storage_group')
    def test_remove_last_vol_and_delete_sg(self, mock_vol_sg,
                                           mock_delete_sg, mock_rm, mock_add):
        self.mask._remove_last_vol_and_delete_sg(
            self.data.array, self.device_id, self.volume_name,
            self.data.storagegroup_name_i, self.extra_specs)
        self.mask._remove_last_vol_and_delete_sg(
            self.data.array, self.device_id, self.volume_name,
            self.data.storagegroup_name_i, self.extra_specs,
            self.data.parent_sg_i, True)
        self.assertEqual(2, mock_delete_sg.call_count)
        self.assertEqual(1, mock_vol_sg.call_count)
        self.assertEqual(1, mock_rm.call_count)
        self.assertEqual(1, mock_add.call_count)

    @mock.patch.object(rest.PowerMaxRest, 'delete_initiator_group')
    def test_last_volume_delete_initiator_group(self, mock_delete_ig):
        self.mask._last_volume_delete_initiator_group(
            self.data.array, self.data.initiatorgroup_name_f, 'Wrong_Host')
        mock_delete_ig.assert_not_called()
        self.mask._last_volume_delete_initiator_group(
            self.data.array, self.data.initiatorgroup_name_f, None)
        mock_delete_ig.assert_not_called()
        mv_list = [self.data.masking_view_name_i,
                   self.data.masking_view_name_f]
        with mock.patch.object(
                rest.PowerMaxRest, 'get_masking_views_by_initiator_group',
                side_effect=[mv_list, []]):
            self.mask._last_volume_delete_initiator_group(
                self.data.array, self.data.initiatorgroup_name_i,
                self.data.connector['host'])
            mock_delete_ig.assert_not_called()
            self.mask._last_volume_delete_initiator_group(
                self.data.array, self.data.initiatorgroup_name_i,
                self.data.connector['host'])
            mock_delete_ig.assert_called_once()

    def test_populate_masking_dict_init_check_false(self):
        extra_specs = deepcopy(self.data.extra_specs)
        extra_specs[utils.PORTGROUPNAME] = self.data.port_group_name_f
        connector = self.data.connector
        with mock.patch.object(self.driver, '_get_initiator_check_flag',
                               return_value=False):
            masking_view_dict = self.driver._populate_masking_dict(
                self.data.test_volume, connector, extra_specs)
            self.assertFalse(masking_view_dict['initiator_check'])

    def test_populate_masking_dict_init_check_true(self):
        extra_specs = deepcopy(self.data.extra_specs)
        extra_specs[utils.PORTGROUPNAME] = self.data.port_group_name_f
        connector = self.data.connector
        with mock.patch.object(self.driver, '_get_initiator_check_flag',
                               return_value=True):
            masking_view_dict = self.driver._populate_masking_dict(
                self.data.test_volume, connector, extra_specs)
            self.assertTrue(masking_view_dict['initiator_check'])

    def test_check_existing_initiator_group_verify_true(self):
        mv_dict = deepcopy(self.data.masking_view_dict)
        mv_dict['initiator_check'] = True
        with mock.patch.object(
                rest.PowerMaxRest, 'get_element_from_masking_view',
                return_value=tpd.PowerMaxData.initiatorgroup_name_f):
            with mock.patch.object(
                    self.mask, '_verify_initiator_group_from_masking_view',
                    return_value=(
                        True,
                        self.data.initiatorgroup_name_f)) as mock_verify:
                self.mask._check_existing_initiator_group(
                    self.data.array, self.data.masking_view_name_f,
                    mv_dict, self.data.storagegroup_name_f,
                    self.data.port_group_name_f, self.data.extra_specs)
                mock_verify.assert_called_once_with(
                    self.data.array, self.data.masking_view_name_f,
                    mv_dict, self.data.initiatorgroup_name_f,
                    self.data.storagegroup_name_f,
                    self.data.port_group_name_f, self.data.extra_specs)

    @mock.patch.object(
        masking.PowerMaxMasking, 'add_child_sg_to_parent_sg',
        side_effect=[None, exception.VolumeBackendAPIException])
    @mock.patch.object(rest.PowerMaxRest, 'is_child_sg_in_parent_sg',
                       side_effect=[True, False, False])
    def test_check_add_child_sg_to_parent_sg(self, mock_is_child, mock_add):
        for x in range(0, 3):
            message = self.mask._check_add_child_sg_to_parent_sg(
                self.data.array, self.data.storagegroup_name_i,
                self.data.parent_sg_i, self.data.extra_specs)
        self.assertIsNotNone(message)

    @mock.patch.object(rest.PowerMaxRest, 'add_child_sg_to_parent_sg')
    @mock.patch.object(rest.PowerMaxRest, 'is_child_sg_in_parent_sg',
                       side_effect=[True, False])
    def test_add_child_sg_to_parent_sg(self, mock_is_child, mock_add):
        for x in range(0, 2):
            self.mask.add_child_sg_to_parent_sg(
                self.data.array, self.data.storagegroup_name_i,
                self.data.parent_sg_i, self.data.extra_specs)
        self.assertEqual(1, mock_add.call_count)

    def test_get_parent_sg_from_child(self):
        with mock.patch.object(self.driver.rest, 'get_storage_group',
                               side_effect=[None, self.data.sg_details[1]]):
            sg_name = self.mask.get_parent_sg_from_child(
                self.data.array, self.data.storagegroup_name_i)
            self.assertIsNone(sg_name)
            sg_name2 = self.mask.get_parent_sg_from_child(
                self.data.array, self.data.storagegroup_name_f)
            self.assertEqual(self.data.parent_sg_f, sg_name2)

    @mock.patch.object(rest.PowerMaxRest, 'get_element_from_masking_view',
                       return_value='parent_sg')
    @mock.patch.object(rest.PowerMaxRest, 'get_num_vols_in_sg',
                       return_value=2)
    def test_get_num_vols_from_mv(self, mock_num, mock_element):
        num_vols, sg = self.mask._get_num_vols_from_mv(
            self.data.array, self.data.masking_view_name_f)
        self.assertEqual(2, num_vols)

    @mock.patch.object(masking.PowerMaxMasking,
                       'add_volume_to_default_storage_group')
    @mock.patch.object(rest.PowerMaxRest, 'delete_storage_group')
    def test_delete_cascaded(self, mock_delete, mock_add):
        self.mask._delete_cascaded_storage_groups(
            self.data.array, self.data.masking_view_name_f,
            self.data.parent_sg_f, self.data.extra_specs,
            self.data.device_id, False)
        self.assertEqual(2, mock_delete.call_count)
        mock_add.assert_not_called()
        # Delete legacy masking view, parent sg = child sg
        mock_delete.reset_mock()
        self.mask._delete_cascaded_storage_groups(
            self.data.array, self.data.masking_view_name_f,
            self.data.masking_view_name_f, self.data.extra_specs,
            self.data.device_id, True)
        self.assertEqual(1, mock_delete.call_count)
        mock_add.assert_called_once()

    @mock.patch.object(masking.PowerMaxMasking,
                       'add_volumes_to_storage_group')
    def test_add_remote_vols_to_volume_group(self, mock_add):
        self.mask.add_remote_vols_to_volume_group(
            [self.data.test_volume], self.data.test_rep_group,
            self.data.rep_extra_specs)
        mock_add.assert_called_once()

    @mock.patch.object(masking.PowerMaxMasking,
                       'add_remote_vols_to_volume_group')
    @mock.patch.object(masking.PowerMaxMasking,
                       '_check_adding_volume_to_storage_group')
    @mock.patch.object(volume_utils, 'is_group_a_cg_snapshot_type',
                       return_value=True)
    @mock.patch.object(volume_utils, 'is_group_a_type',
                       side_effect=[False, False, True, True])
    def test_return_volume_to_volume_group(self, mock_type, mock_cg,
                                           mock_check, mock_add):
        vol_grp_member = deepcopy(self.data.test_volume)
        vol_grp_member.group_id = self.data.test_vol_grp_name_id_only
        vol_grp_member.group = self.data.test_group
        for x in range(0, 2):
            self.mask.return_volume_to_volume_group(
                self.data.array, vol_grp_member, self.data.device_id,
                self.data.test_volume.name, self.data.extra_specs)
        mock_add.assert_called_once()

    @mock.patch.object(masking.PowerMaxMasking,
                       '_return_volume_to_fast_managed_group')
    def test_pre_multiattach(self, mock_return):
        mv_dict = self.mask.pre_multiattach(
            self.data.array, self.data.device_id,
            self.data.masking_view_dict_multiattach,
            self.data.extra_specs)
        mock_return.assert_not_called()
        self.assertEqual(self.data.storagegroup_name_f,
                         mv_dict[utils.FAST_SG])
        with mock.patch.object(
                self.mask, 'move_volume_between_storage_groups',
                side_effect=exception.CinderException):
            self.assertRaises(
                exception.VolumeBackendAPIException,
                self.mask.pre_multiattach, self.data.array,
                self.data.device_id, self.data.masking_view_dict_multiattach,
                self.data.extra_specs)
            mock_return.assert_called_once()

    def test_pre_multiattach_next_gen(self):
        with mock.patch.object(utils.PowerMaxUtils, 'truncate_string',
                               return_value='DiamondDSS'):
            self.mask.pre_multiattach(
                self.data.array, self.data.device_id,
                self.data.masking_view_dict_multiattach,
                self.data.extra_specs)
            utils.PowerMaxUtils.truncate_string.assert_called_once_with(
                'DiamondDSS', 10)

    @mock.patch.object(masking.PowerMaxMasking,
                       '_clean_up_child_storage_group')
    @mock.patch.object(masking.PowerMaxMasking,
                       'move_volume_between_storage_groups')
    @mock.patch.object(masking.PowerMaxMasking,
                       '_return_volume_to_fast_managed_group')
    def test_pre_multiattach_pool_none_workload(self, mock_return, mck_move,
                                                mck_clean):
        with mock.patch.object(utils.PowerMaxUtils, 'truncate_string',
                               return_value='OptimdNONE'):
            self.mask.pre_multiattach(
                self.data.array, self.data.device_id,
                self.data.masking_view_dict_multiattach,
                self.data.extra_specs_optimized)
            utils.PowerMaxUtils.truncate_string.assert_called_once_with(
                'OptimizedNONE', 10)

    @mock.patch.object(
        rest.PowerMaxRest, 'get_storage_group_list',
        side_effect=[
            {'storageGroupId': [tpd.PowerMaxData.no_slo_sg_name]}, {}])
    @mock.patch.object(masking.PowerMaxMasking,
                       '_return_volume_to_fast_managed_group')
    def test_check_return_volume_to_fast_managed_group(
            self, mock_return, mock_sg):
        for x in range(0, 2):
            self.mask.return_volume_to_fast_managed_group(
                self.data.array, self.data.device_id,
                self.data.extra_specs)
        no_slo_specs = deepcopy(self.data.extra_specs)
        no_slo_specs[utils.SLO] = None
        self.mask.return_volume_to_fast_managed_group(
            self.data.array, self.data.device_id, no_slo_specs)
        mock_return.assert_called_once()

    @mock.patch.object(masking.PowerMaxMasking, '_move_vol_from_default_sg')
    @mock.patch.object(masking.PowerMaxMasking,
                       '_clean_up_child_storage_group')
    @mock.patch.object(masking.PowerMaxMasking, 'add_child_sg_to_parent_sg')
    @mock.patch.object(masking.PowerMaxMasking, '_get_or_create_storage_group')
    @mock.patch.object(
        rest.PowerMaxRest, 'get_storage_groups_from_volume',
        side_effect=[[tpd.PowerMaxData.no_slo_sg_name],
                     [tpd.PowerMaxData.storagegroup_name_f]])
    def test_return_volume_to_fast_managed_group(
            self, mock_sg, mock_get, mock_add, mock_clean, mock_move):
        for x in range(0, 2):
            self.mask._return_volume_to_fast_managed_group(
                self.data.array, self.data.device_id,
                self.data.parent_sg_f, self.data.storagegroup_name_f,
                self.data.no_slo_sg_name, self.data.extra_specs)
        mock_get.assert_called_once()
        mock_clean.assert_called_once()

    @mock.patch.object(rest.PowerMaxRest, 'delete_storage_group')
    @mock.patch.object(rest.PowerMaxRest, 'remove_child_sg_from_parent_sg')
    @mock.patch.object(rest.PowerMaxRest, 'is_child_sg_in_parent_sg',
                       side_effect=[False, True])
    @mock.patch.object(rest.PowerMaxRest, 'get_num_vols_in_sg',
                       side_effect=[2, 0, 0])
    @mock.patch.object(rest.PowerMaxRest, 'get_storage_group', side_effect=[
        None, 'child_sg', 'child_sg', 'child_sg'])
    def test_clean_up_child_storage_group(
            self, mock_sg, mock_num, mock_child, mock_rm, mock_del):
        # Storage group not found
        self.mask._clean_up_child_storage_group(
            self.data.array, self.data.storagegroup_name_f,
            self.data.parent_sg_f, self.data.extra_specs)
        mock_num.assert_not_called()
        # Storage group not empty
        self.mask._clean_up_child_storage_group(
            self.data.array, self.data.storagegroup_name_f,
            self.data.parent_sg_f, self.data.extra_specs)
        mock_child.assert_not_called()
        # Storage group not child
        self.mask._clean_up_child_storage_group(
            self.data.array, self.data.storagegroup_name_f,
            self.data.parent_sg_f, self.data.extra_specs)
        mock_rm.assert_not_called()
        # Storage group is child, and empty
        self.mask._clean_up_child_storage_group(
            self.data.array, self.data.storagegroup_name_f,
            self.data.parent_sg_f, self.data.extra_specs)
        mock_rm.assert_called_once()
        self.assertEqual(2, mock_del.call_count)

    @mock.patch.object(utils.PowerMaxUtils, 'verify_tag_list')
    def test_add_tags_to_storage_group_disabled(self, mock_verify):
        self.mask._add_tags_to_storage_group(
            self.data.array, self.data.add_volume_sg_info_dict,
            self.data.extra_specs)
        mock_verify.assert_not_called()

    @mock.patch.object(utils.PowerMaxUtils, 'verify_tag_list')
    def test_add_tags_to_storage_group_enabled(self, mock_verify):
        self.mask._add_tags_to_storage_group(
            self.data.array, self.data.add_volume_sg_info_dict,
            self.data.extra_specs_tags)
        mock_verify.assert_called()

    @mock.patch.object(utils.PowerMaxUtils, 'get_new_tags')
    def test_add_tags_to_storage_group_existing_tags(self, mock_inter):
        self.mask._add_tags_to_storage_group(
            self.data.array, self.data.storage_group_with_tags,
            self.data.extra_specs_tags)
        mock_inter.assert_called()

    @mock.patch.object(rest.PowerMaxRest, 'add_storage_group_tag',
                       side_effect=[exception.VolumeBackendAPIException])
    def test_add_tags_to_storage_group_exception(self, mock_except):
        self.mask._add_tags_to_storage_group(
            self.data.array, self.data.add_volume_sg_info_dict,
            self.data.extra_specs_tags)
        mock_except.assert_called()

    @mock.patch.object(rest.PowerMaxRest,
                       'get_masking_views_from_storage_group',
                       return_value=[tpd.PowerMaxData.masking_view_name_f])
    def test_get_host_and_port_group_labels(self, mock_mv):
        host_label, port_group_label = (
            self.mask._get_host_and_port_group_labels(
                self.data.array, self.data.parent_sg_f))
        self.assertEqual('HostX', host_label)
        self.assertEqual('OS-fibre-PG', port_group_label)

    @mock.patch.object(rest.PowerMaxRest,
                       'get_masking_views_from_storage_group',
                       return_value=['OS-HostX699ea-I-p-name3b02c-MV'])
    def test_get_host_and_port_group_labels_complex(self, mock_mv):
        host_label, port_group_label = (
            self.mask._get_host_and_port_group_labels(
                self.data.array, self.data.parent_sg_f))
        self.assertEqual('HostX699ea', host_label)
        self.assertEqual('p-name3b02c', port_group_label)

    @mock.patch.object(rest.PowerMaxRest,
                       'get_masking_views_from_storage_group',
                       return_value=['OS-myhost-I-myportgroup-MV'])
    def test_get_host_and_port_group_labels_plain(self, mock_mv):
        host_label, port_group_label = (
            self.mask._get_host_and_port_group_labels(
                self.data.array, self.data.parent_sg_f))
        self.assertEqual('myhost', host_label)
        self.assertEqual('myportgroup', port_group_label)

    @mock.patch.object(rest.PowerMaxRest,
                       'get_masking_views_from_storage_group',
                       return_value=[
                           'OS-host-with-dash-I-portgroup-with-dashes-MV'])
    def test_get_host_and_port_group_labels_dashes(self, mock_mv):
        host_label, port_group_label = (
            self.mask._get_host_and_port_group_labels(
                self.data.array, self.data.parent_sg_f))
        self.assertEqual('host-with-dash', host_label)
        self.assertEqual('portgroup-with-dashes', port_group_label)

    @mock.patch.object(
        rest.PowerMaxRest, 'is_child_sg_in_parent_sg', return_value=False)
    @mock.patch.object(
        rest.PowerMaxRest, 'add_child_sg_to_parent_sg')
    @mock.patch.object(
        rest.PowerMaxRest, 'get_storage_group',
        side_effect=[None, tpd.PowerMaxData.sg_details[1],
                     tpd.PowerMaxData.sg_details[2]])
    @mock.patch.object(
        provision.PowerMaxProvision, 'create_storage_group')
    def test_check_child_storage_group_exists_false(
            self, mock_create, mock_get, mock_add, mock_check):
        self.mask._check_child_storage_group_exists(
            self.data.device_id, self.data.array,
            self.data.storagegroup_name_i, self.data.extra_specs,
            self.data.parent_sg_i)
        mock_create.assert_called_once()
        mock_add.assert_called_once()

    @mock.patch.object(
        rest.PowerMaxRest, 'is_child_sg_in_parent_sg', return_value=True)
    @mock.patch.object(
        rest.PowerMaxRest, 'add_child_sg_to_parent_sg')
    @mock.patch.object(
        rest.PowerMaxRest, 'get_storage_group',
        side_effect=[tpd.PowerMaxData.sg_details[1],
                     tpd.PowerMaxData.sg_details[3]])
    @mock.patch.object(
        provision.PowerMaxProvision, 'create_storage_group')
    def test_check_child_storage_group_exists_true(
            self, mock_create, mock_get, mock_add, mock_check):
        self.mask._check_child_storage_group_exists(
            self.data.device_id, self.data.array,
            self.data.storagegroup_name_i, self.data.extra_specs,
            self.data.parent_sg_i)
        mock_create.assert_not_called
        mock_add.assert_not_called()

    @mock.patch.object(rest.PowerMaxRest, 'get_port',
                       return_value=tpd.PowerMaxData.port_info)
    @mock.patch.object(rest.PowerMaxRest, 'get_port_ids',
                       return_value=['FA-1D:4'])
    def test_check_director_and_port_status(self, mock_port_ids, mock_port):
        self.mask._check_director_and_port_status(
            self.data.array, self.data.port_group_name_f)

    @mock.patch.object(rest.PowerMaxRest, 'get_port',
                       return_value=tpd.PowerMaxData.port_info_off)
    @mock.patch.object(rest.PowerMaxRest, 'get_port_ids',
                       return_value=['FA-1D:4'])
    def test_check_director_and_port_status_invalid_status(
            self, mock_port_ids, mock_port):
        exception_message = (
            r"The director status is Offline and the port status is OFF for "
            r"dir:port FA-1D:4.")

        with self.assertRaisesRegex(
                exception.VolumeBackendAPIException,
                exception_message):
            self.mask._check_director_and_port_status(
                self.data.array, self.data.port_group_name_f)

    @mock.patch.object(rest.PowerMaxRest, 'get_port',
                       return_value=tpd.PowerMaxData.port_info_no_status)
    @mock.patch.object(rest.PowerMaxRest, 'get_port_ids',
                       return_value=['FA-1D:4'])
    def test_check_director_and_port_status_no_status(
            self, mock_port_ids, mock_port):
        exception_message = (
            r"Unable to get the director or port status for dir:port "
            r"FA-1D:4.")

        with self.assertRaisesRegex(
                exception.VolumeBackendAPIException,
                exception_message):
            self.mask._check_director_and_port_status(
                self.data.array, self.data.port_group_name_f)

    @mock.patch.object(rest.PowerMaxRest, 'get_port',
                       return_value=tpd.PowerMaxData.port_info_no_details)
    @mock.patch.object(rest.PowerMaxRest, 'get_port_ids',
                       return_value=['FA-1D:4'])
    def test_check_director_and_port_status_no_details(
            self, mock_port_ids, mock_port):
        exception_message = (
            r"Unable to get port information for dir:port FA-1D:4.")

        with self.assertRaisesRegex(
                exception.VolumeBackendAPIException,
                exception_message):
            self.mask._check_director_and_port_status(
                self.data.array, self.data.port_group_name_f)

    @mock.patch.object(
        rest.PowerMaxRest, 'get_storage_groups_from_volume',
        return_value=([tpd.PowerMaxData.storagegroup_name_f]))
    @mock.patch.object(
        rest.PowerMaxRest, 'get_masking_views_from_storage_group',
        return_value=[tpd.PowerMaxData.masking_view_name_f])
    def test_validate_attach(self, mock_sgs, mock_mvs):
        self.assertTrue(self.mask._validate_attach(
            self.data.array, self.data.device_id,
            self.data.storagegroup_name_f, self.data.masking_view_name_f))

    @mock.patch.object(
        rest.PowerMaxRest, 'get_storage_groups_from_volume',
        return_value=([tpd.PowerMaxData.storagegroup_name_f]))
    @mock.patch.object(
        rest.PowerMaxRest, 'get_masking_views_from_storage_group',
        return_value=[])
    def test_validate_attach_no_mvs(self, mock_sgs, mock_mvs):
        self.assertFalse(self.mask._validate_attach(
            self.data.array, self.data.device_id,
            self.data.storagegroup_name_f, self.data.masking_view_name_f))

    @mock.patch.object(
        rest.PowerMaxRest, 'get_storage_groups_from_volume',
        return_value=([tpd.PowerMaxData.defaultstoragegroup_name]))
    @mock.patch.object(
        rest.PowerMaxRest, 'get_masking_views_from_storage_group',
        return_value=[])
    def test_validate_attach_incorrect_sg(self, mock_sgs, mock_mvs):
        self.assertFalse(self.mask._validate_attach(
            self.data.array, self.data.device_id,
            self.data.storagegroup_name_f, self.data.masking_view_name_f))



# Copyright (C) 2017 HuaWei Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Tests for attachments Api."""

from unittest import mock

import ddt

from cinder.api import microversions as mv
from cinder.api.v3 import attachments as v3_attachments
from cinder import context
from cinder import exception
from cinder import objects
from cinder.tests.unit.api import fakes
from cinder.tests.unit import fake_constants as fake
from cinder.tests.unit import test
from cinder.volume import api as volume_api
from cinder.volume import rpcapi as volume_rpcapi
from cinder.volume import volume_utils


@ddt.ddt
class AttachmentsAPITestCase(test.TestCase):
    """Test Case for attachment API."""

    def setUp(self):
        super(AttachmentsAPITestCase, self).setUp()
        self.controller = v3_attachments.AttachmentsController()
        self.volume_api = volume_api.API()
        self.ctxt = context.RequestContext(fake.USER_ID, fake.PROJECT_ID,
                                           auth_token=True,
                                           is_admin=True)
        self.volume1 = self._create_volume(display_name='fake_volume_1',
                                           project_id=fake.PROJECT_ID)
        self.volume2 = self._create_volume(display_name='fake_volume_2',
                                           project_id=fake.PROJECT2_ID)
        self.attachment1 = self._create_attachment(
            volume_uuid=self.volume1.id, instance_uuid=fake.UUID1,
            host='host-a')
        self.attachment2 = self._create_attachment(
            volume_uuid=self.volume1.id, instance_uuid=fake.UUID1,
            host='host-b')
        self.attachment3 = self._create_attachment(
            volume_uuid=self.volume1.id, instance_uuid=fake.UUID2,
            host='host-c')
        self.attachment4 = self._create_attachment(
            volume_uuid=self.volume2.id, instance_uuid=fake.UUID2,
            host='host-d')
        self.addCleanup(self._cleanup)

    def _cleanup(self):
        self.attachment1.destroy()
        self.attachment2.destroy()
        self.attachment3.destroy()
        self.attachment4.destroy()
        self.volume1.destroy()
        self.volume2.destroy()

    def _create_volume(self, ctxt=None, display_name=None, project_id=None):
        """Create a volume object."""
        ctxt = ctxt or self.ctxt
        volume = objects.Volume(ctxt)
        volume.display_name = display_name
        volume.project_id = project_id
        volume.status = 'available'
        volume.attach_status = 'attached'
        volume.create()
        return volume

    def test_create_attachment(self):
        req = fakes.HTTPRequest.blank('/v3/%s/attachments' %
                                      fake.PROJECT_ID,
                                      version=mv.NEW_ATTACH)
        body = {
            "attachment":
                {
                    "connector": None,
                    "instance_uuid": fake.UUID1,
                    "volume_uuid": self.volume1.id
                },
        }

        attachment = self.controller.create(req, body=body)

        self.assertEqual(self.volume1.id,
                         attachment['attachment']['volume_id'])
        self.assertEqual(fake.UUID1,
                         attachment['attachment']['instance'])

    @mock.patch.object(volume_rpcapi.VolumeAPI, 'attachment_update')
    def test_update_attachment(self, mock_update):
        fake_connector = {'fake_key': 'fake_value',
                          'host': 'somehost'}
        mock_update.return_value = fake_connector
        req = fakes.HTTPRequest.blank('/v3/%s/attachments/%s' %
                                      (fake.PROJECT_ID, self.attachment1.id),
                                      version=mv.NEW_ATTACH,
                                      use_admin_context=True)
        body = {
            "attachment":
                {
                    "connector": {'fake_key': 'fake_value',
                                  'host': 'somehost',
                                  'connection_info': 'a'},
                },
        }

        attachment = self.controller.update(req, self.attachment1.id,
                                            body=body)

        self.assertEqual(fake_connector,
                         attachment['attachment']['connection_info'])
        self.assertEqual(fake.UUID1, attachment['attachment']['instance'])

    def test_update_attachment_with_empty_connector_object(self):
        req = fakes.HTTPRequest.blank('/v3/%s/attachments/%s' %
                                      (fake.PROJECT_ID, self.attachment1.id),
                                      version=mv.NEW_ATTACH,
                                      use_admin_context=True)
        body = {
            "attachment":
                {
                    "connector": {},
                },
        }
        self.assertRaises(exception.ValidationError,
                          self.controller.update, req,
                          self.attachment1.id, body=body)

    @ddt.data(mv.get_prior_version(mv.RESOURCE_FILTER),
              mv.RESOURCE_FILTER, mv.LIKE_FILTER)
    @mock.patch('cinder.api.common.reject_invalid_filters')
    def test_attachment_list_with_general_filter(self, version, mock_update):
        url = '/v3/%s/attachments' % fake.PROJECT_ID
        req = fakes.HTTPRequest.blank(url,
                                      version=version,
                                      use_admin_context=False)
        self.controller.index(req)

        if version != mv.get_prior_version(mv.RESOURCE_FILTER):
            support_like = True if version == mv.LIKE_FILTER else False
            mock_update.assert_called_once_with(req.environ['cinder.context'],
                                                mock.ANY, 'attachment',
                                                support_like)

    @ddt.data('reserved', 'attached')
    @mock.patch.object(volume_rpcapi.VolumeAPI, 'attachment_delete')
    def test_delete_attachment(self, status, mock_delete):
        volume1 = self._create_volume(display_name='fake_volume_1',
                                      project_id=fake.PROJECT_ID)
        attachment = self._create_attachment(
            volume_uuid=volume1.id, instance_uuid=fake.UUID1,
            attach_status=status)
        req = fakes.HTTPRequest.blank('/v3/%s/attachments/%s' %
                                      (fake.PROJECT_ID, attachment.id),
                                      version=mv.NEW_ATTACH,
                                      use_admin_context=True)

        self.controller.delete(req, attachment.id)

        volume2 = objects.Volume.get_by_id(self.ctxt, volume1.id)
        # Volume and attachment status is changed on the API service
        self.assertEqual('detached', volume2.attach_status)
        self.assertEqual('available', volume2.status)
        self.assertRaises(
            exception.VolumeAttachmentNotFound,
            objects.VolumeAttachment.get_by_id, self.ctxt, attachment.id)
        if status != 'reserved':
            mock_delete.assert_called_once_with(req.environ['cinder.context'],
                                                attachment.id, mock.ANY)

    def _create_attachment(self, ctxt=None, volume_uuid=None,
                           instance_uuid=None, mountpoint=None,
                           attach_time=None, detach_time=None,
                           attach_status=None, attach_mode=None, host=''):
        """Create an attachment object."""
        ctxt = ctxt or self.ctxt
        attachment = objects.VolumeAttachment(ctxt)
        attachment.volume_id = volume_uuid
        attachment.instance_uuid = instance_uuid
        attachment.mountpoint = mountpoint
        attachment.attach_time = attach_time
        attachment.detach_time = detach_time
        attachment.attach_status = attach_status or 'reserved'
        attachment.attach_mode = attach_mode
        attachment.connector = {'host': host}
        attachment.create()
        return attachment

    @ddt.data("instance_uuid", "volume_uuid")
    def test_create_attachment_without_resource_uuid(self, resource_uuid):
        req = fakes.HTTPRequest.blank('/v3/%s/attachments' %
                                      fake.PROJECT_ID,
                                      version=mv.NEW_ATTACH)
        body = {
            "attachment":
                {
                    "connector": None
                }
        }
        body["attachment"][resource_uuid] = "test_id"

        self.assertRaises(exception.ValidationError,
                          self.controller.create, req, body=body)

    @ddt.data(
        {"attachment": {
            "connector": None,
            "instance_uuid": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa",
            "volume_uuid": "invalid-uuid"}},
        {"attachment": {
            "instance_uuid": "invalid-uuid",
            "volume_uuid": "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"}})
    def test_create_attachment_with_invalid_resource_uuid(self, fake_body):
        req = fakes.HTTPRequest.blank('/v3/%s/attachments' %
                                      fake.PROJECT_ID,
                                      version=mv.NEW_ATTACH)

        self.assertRaises(exception.ValidationError,
                          self.controller.create, req, body=fake_body)

    @mock.patch('cinder.volume.api.API._attachment_reserve')
    def test_create_attachment_in_use_volume_multiattach_false(self,
                                                               mock_reserve):
        """Negative test for creating an attachment on an in-use volume."""
        req = fakes.HTTPRequest.blank('/v3/%s/attachments' %
                                      fake.PROJECT_ID,
                                      version=mv.NEW_ATTACH)
        body = {
            "attachment":
                {
                    "connector": None,
                    "instance_uuid": fake.UUID1,
                    "volume_uuid": self.volume1.id
                },
        }
        mock_reserve.side_effect = (
            exception.InvalidVolume(
                reason="Volume %s status must be available or "
                       "downloading" % self.volume1.id))
        # Note that if we were using the full WSGi stack, the
        # ResourceExceptionHandler would convert this to an HTTPBadRequest.
        self.assertRaises(exception.InvalidVolume,
                          self.controller.create, req, body=body)

    @ddt.data(False, True)
    def test_list_attachments(self, is_detail):
        url = '/v3/%s/attachments' % fake.PROJECT_ID
        list_func = self.controller.index
        if is_detail:
            url = '/v3/%s/groups/detail' % fake.PROJECT_ID
            list_func = self.controller.detail
        req = fakes.HTTPRequest.blank(url, version=mv.NEW_ATTACH,
                                      use_admin_context=True)
        res_dict = list_func(req)

        self.assertEqual(1, len(res_dict))
        self.assertEqual(3, len(res_dict['attachments']))
        self.assertEqual(self.attachment3.id,
                         res_dict['attachments'][0]['id'])

    def test_list_attachments_with_limit(self):
        url = '/v3/%s/attachments?limit=1' % fake.PROJECT_ID
        req = fakes.HTTPRequest.blank(url, version=mv.NEW_ATTACH,
                                      use_admin_context=True)
        res_dict = self.controller.index(req)

        self.assertEqual(1, len(res_dict))
        self.assertEqual(1, len(res_dict['attachments']))

    def test_list_attachments_with_marker(self):
        url = '/v3/%s/attachments?marker=%s' % (fake.PROJECT_ID,
                                                self.attachment3.id)
        req = fakes.HTTPRequest.blank(url, version=mv.NEW_ATTACH,
                                      use_admin_context=True)
        res_dict = self.controller.index(req)

        self.assertEqual(1, len(res_dict))
        self.assertEqual(2, len(res_dict['attachments']))
        self.assertEqual(self.attachment2.id,
                         res_dict['attachments'][0]['id'])

    @ddt.data("desc", "asc")
    def test_list_attachments_with_sort(self, sort_dir):
        url = '/v3/%s/attachments?sort_key=id&sort_dir=%s' % (fake.PROJECT_ID,
                                                              sort_dir)
        req = fakes.HTTPRequest.blank(url, version=mv.NEW_ATTACH,
                                      use_admin_context=True)
        res_dict = self.controller.index(req)

        self.assertEqual(1, len(res_dict))
        self.assertEqual(3, len(res_dict['attachments']))
        order_ids = sorted([self.attachment1.id,
                            self.attachment2.id,
                            self.attachment3.id])
        expect_result = order_ids[2] if sort_dir == "desc" else order_ids[0]
        self.assertEqual(expect_result,
                         res_dict['attachments'][0]['id'])

    @ddt.data({'admin': True, 'request_url': '?all_tenants=1', 'count': 4},
              {'admin': False, 'request_url': '?all_tenants=1', 'count': 3},
              {'admin': True, 'request_url':
                  '?all_tenants=1&project_id=%s' % fake.PROJECT2_ID,
               'count': 1},
              {'admin': False, 'request_url': '', 'count': 3},
              {'admin': False, 'request_url': '?instance_id=%s' % fake.UUID1,
               'count': 2},
              {'admin': False, 'request_url': '?instance_id=%s' % fake.UUID2,
               'count': 1})
    @ddt.unpack
    def test_list_attachment_with_tenants(self, admin, request_url, count):
        url = '/v3/%s/attachments%s' % (fake.PROJECT_ID, request_url)
        req = fakes.HTTPRequest.blank(url, version=mv.NEW_ATTACH,
                                      use_admin_context=admin)
        res_dict = self.controller.index(req)

        self.assertEqual(1, len(res_dict))
        self.assertEqual(count, len(res_dict['attachments']))

    @mock.patch.object(volume_utils, 'notify_about_volume_usage')
    def test_complete_attachment(self, mock_notify):

        def fake_notify(context, volume, event_suffix,
                        extra_usage_info=None, host=None):
            # Check the notify content is in-use volume and 'attach.end'
            self.assertEqual('in-use', volume['status'])
            self.assertEqual('attach.end', event_suffix)

        mock_notify.side_effect = fake_notify
        req = fakes.HTTPRequest.blank('/v3/%s/attachments/%s/action' %
                                      (fake.PROJECT_ID, self.attachment1.id),
                                      version=mv.NEW_ATTACH_COMPLETION,
                                      use_admin_context=True)
        body = {"os-complete": {}}

        self.controller.complete(req, self.attachment1.id,
                                 body=body)
        # Check notify has been called once
        mock_notify.assert_called_once()



#  Copyright (C) 2021-2022 YADRO.
#  All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License"); you may
#  not use this file except in compliance with the License. You may obtain
#  a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#  WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#  License for the specific language governing permissions and limitations
#  under the License.

import os
from unittest.mock import mock_open
from unittest.mock import patch
from unittest import TestCase

from cinder.volume.drivers.yadro.tatlin_utils import TatlinVolumeConnections

VOL_ID = 'cinder-volume-id'


class TatlinVolumeConnectionsTest(TestCase):

    @patch('oslo_concurrency.lockutils.lock', autospec=True)
    @patch('os.mkdir')
    @patch('os.path.isdir')
    def setUp(self, isdir, mkdir, lock):
        self.path = 'fake_path'
        isdir.return_value = False
        self.connections = TatlinVolumeConnections(self.path)
        isdir.assert_called_once_with(self.path)
        mkdir.assert_called_once_with(self.path)
        isdir.reset_mock()
        mkdir.reset_mock()
        isdir.return_value = True
        self.connections = TatlinVolumeConnections(self.path)
        isdir.assert_called_once_with(self.path)
        mkdir.assert_not_called()

    @patch('oslo_concurrency.lockutils.lock', autospec=True)
    @patch('builtins.open', mock_open(read_data='1'))
    @patch('os.path.exists')
    def test_get(self, exists, lock):
        exists.side_effect = [False, True]
        self.assertEqual(self.connections.get(VOL_ID), 0)
        self.assertEqual(self.connections.get(VOL_ID), 1)

    @patch('oslo_concurrency.lockutils.lock', autospec=True)
    @patch('builtins.open', callable=mock_open(read_data='1'))
    @patch('os.path.exists')
    def test_increment(self, exists, open, lock):
        exists.side_effect = [False, True]
        self.assertEqual(self.connections.increment(VOL_ID), 1)
        open.assert_called_once_with(os.path.join(self.path, VOL_ID), 'w')
        with open() as f:
            f.write.assert_called_once_with('1')
        self.assertEqual(self.connections.increment(VOL_ID), 2)
        open.assert_called_with(os.path.join(self.path, VOL_ID), 'w')
        with open() as f:
            f.write.assert_called_with('2')

    @patch('oslo_concurrency.lockutils.lock', autospec=True)
    @patch('builtins.open', callable=mock_open())
    @patch('os.remove')
    @patch('os.path.exists')
    def test_decrement(self, exists, remove, open, lock):
        exists.side_effect = [False, True, True]
        with open() as f:
            f.read.side_effect = [2, 1]

            self.assertEqual(self.connections.decrement(VOL_ID), 0)
            remove.assert_not_called()

            self.assertEqual(self.connections.decrement(VOL_ID), 1)
            open.assert_called_with(os.path.join(self.path, VOL_ID), 'w')
            f.write.assert_called_with('1')

            self.assertEqual(self.connections.decrement(VOL_ID), 0)
            remove.assert_called_with(os.path.join(self.path, VOL_ID))



# Copyright (C) 2016-2022 Lightbits Labs Ltd.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from copy import deepcopy
import functools
import hashlib
import http.client as httpstatus
import json
from typing import Dict
from typing import List
from typing import Tuple
from unittest import mock
import uuid

from cinder import context
from cinder import db
from cinder import exception
from cinder.tests.unit import test
from cinder.tests.unit import utils as test_utils
from cinder.volume import configuration as conf
from cinder.volume.drivers import lightos


FAKE_LIGHTOS_CLUSTER_NODES: Dict[str, List] = {
    "nodes": [
        {"UUID": "926e6df8-73e1-11ec-a624-000000000001",
         "nvmeEndpoint": "192.168.75.10:4420"},
        {"UUID": "926e6df8-73e1-11ec-a624-000000000002",
         "nvmeEndpoint": "192.168.75.11:4420"},
        {"UUID": "926e6df8-73e1-11ec-a624-000000000003",
         "nvmeEndpoint": "192.168.75.12:4420"}
    ]
}
IPV6_LIST = ['::192:168:75:10', '::192:168:75:11', '::192:168:75:12']
FAKE_LIGHTOS_CLUSTER_NODES_IPV6: Dict[str, List] = {
    "nodes": [
        {"UUID": "926e6df8-73e1-11ec-a624-000000000001",
         "nvmeEndpoint": "[{}]:4420".format(IPV6_LIST[0])},
        {"UUID": "926e6df8-73e1-11ec-a624-000000000002",
         "nvmeEndpoint": "[{}]:4420".format(IPV6_LIST[1])},
        {"UUID": "926e6df8-73e1-11ec-a624-000000000003",
         "nvmeEndpoint": "[{}]:4420".format(IPV6_LIST[2])}
    ]
}

FAKE_LIGHTOS_CLUSTER_INFO: Dict[str, str] = {
    'UUID': "926e6df8-73e1-11ec-a624-07ba3880f6cc",
    'subsystemNQN': "nqn.2014-08.org.nvmexpress:NVMf:uuid:"
    "f4a89ce0-9fc2-4900-bfa3-00ad27995e7b"
}

FAKE_CLIENT_HOSTNQN = "hostnqn1"
VOLUME_BACKEND_NAME = "lightos_backend"
RESERVED_PERCENTAGE = 30
DEVICE_SCAN_ATTEMPTS_DEFAULT = 5
LIGHTOS_API_SERVICE_TIMEOUT = 30
VOLUME_BACKEND_NAME = "lightos_backend"
RESERVED_PERCENTAGE = 30
DEFAULT_COMPRESSION = False


class InitiatorConnectorFactoryMocker:
    @staticmethod
    def factory(protocol, root_helper, driver=None,
                use_multipath=False,
                device_scan_attempts=DEVICE_SCAN_ATTEMPTS_DEFAULT,
                arch=None,
                *args, **kwargs):
        return InitialConnectorMock()


class InitialConnectorMock:
    nqn = FAKE_CLIENT_HOSTNQN
    found_discovery_client = True

    def get_hostnqn(self):
        return self.__class__.nqn

    def find_dsc(self):
        return self.__class__.found_discovery_client

    def get_connector_properties(self, root):
        return dict(nqn=self.__class__.nqn,
                    found_dsc=self.__class__.found_discovery_client)


def get_connector_properties():
    connector = InitialConnectorMock()
    return dict(nqn=connector.get_hostnqn(),
                found_dsc=connector.find_dsc())


def get_vol_etag(volume):
    v = deepcopy(volume)
    v.pop("ETag", None)
    dump = json.dumps(v, sort_keys=True).encode('utf-8')
    return hashlib.md5(dump).hexdigest()


class DBMock(object):

    def __init__(self):
        self.data = {
            "projects": {},
        }

    def get_or_create_project(self, project_name) -> Dict:
        project = self.data["projects"].setdefault(project_name, {})
        return project

    def get_project(self, project_name) -> Dict:
        project = self.data["projects"].get(project_name, None)
        return project if project else None

    def delete_project(self, project_name) -> Dict:
        assert project_name != "default", "can't delete default project"
        project = self.get_project(project_name)
        if not project:
            return None
        self.data["projects"].remove(project)
        return project

    def create_volume(self, volume) -> Tuple[int, Dict]:
        assert volume["project_name"] and volume["name"], "must be provided"
        project = self.get_or_create_project(volume["project_name"])
        volumes = project.setdefault("volumes", [])
        existing_volume = next(iter([vol for vol in volumes
                                     if vol["name"] == volume["name"]]), None)
        if not existing_volume:
            volume["UUID"] = str(uuid.uuid4())
            volumes.append(volume)
            return httpstatus.OK, volume
        return httpstatus.CONFLICT, None

    def get_volume_by_uuid(self, project_name,
                           volume_uuid) -> Tuple[int, Dict]:
        assert project_name and volume_uuid, "must be provided"
        project = self.get_project(project_name)
        if not project:
            return httpstatus.NOT_FOUND, None
        proj_vols = project.get("volumes", None)
        if not proj_vols:
            return httpstatus.NOT_FOUND, None
        vol = next(iter([vol for vol in proj_vols
                         if vol["UUID"] == volume_uuid]), None)
        return (httpstatus.OK, vol) if vol else (httpstatus.NOT_FOUND, None)

    def update_volume_by_uuid(self, project_name,
                              volume_uuid, **kwargs) -> Tuple[int, Dict]:
        error_code, volume = self.get_volume_by_uuid(project_name, volume_uuid)
        if error_code != httpstatus.OK:
            return error_code, None
        etag = kwargs.get("etag", None)
        if etag:
            vol_etag = volume.get("ETag", None)
            if etag != vol_etag:
                return httpstatus.BAD_REQUEST, None
        if kwargs.get("size", None):
            volume["size"] = kwargs["size"]
        if kwargs.get("acl", None):
            volume["acl"] = {'values': kwargs.get('acl')}
        volume["ETag"] = get_vol_etag(volume)
        return httpstatus.OK, volume

    def get_volume_by_name(self, project_name,
                           volume_name) -> Tuple[int, Dict]:
        assert project_name and volume_name, "must be provided"
        project = self.get_project(project_name)
        if not project:
            return httpstatus.NOT_FOUND, None
        proj_vols = project.get("volumes", None)
        if not proj_vols:
            return httpstatus.NOT_FOUND, None
        vol = next(iter([vol for vol in proj_vols
                         if vol["name"] == volume_name]), None)
        return (httpstatus.OK, vol) if vol else (httpstatus.NOT_FOUND, None)

    def delete_volume(self, project_name, volume_uuid) -> Tuple[int, Dict]:
        assert project_name and volume_uuid, "must be provided"
        project = self.get_project(project_name)
        if not project:
            return httpstatus.NOT_FOUND, None
        proj_vols = project.get("volumes", None)
        if not proj_vols:
            return httpstatus.NOT_FOUND, None
        for vol in proj_vols:
            if vol["UUID"] == volume_uuid:
                proj_vols.remove(vol)
        return httpstatus.OK, vol

    def update_volume(self, **kwargs):
        assert("project_name" in kwargs and kwargs["project_name"]), \
            "project_name must be provided"

    def create_snapshot(self, snapshot) -> Tuple[int, Dict]:
        assert snapshot["project_name"] and snapshot["name"], \
            "must be provided"
        project = self.get_or_create_project(snapshot["project_name"])
        snapshots = project.setdefault("snapshots", [])
        existing_snap = next(iter([snap for snap in snapshots
                                   if snap["name"] == snapshot["name"]]), None)
        if not existing_snap:
            snapshot["UUID"] = str(uuid.uuid4())
            snapshots.append(snapshot)
            return httpstatus.OK, snapshot
        return httpstatus.CONFLICT, None

    def delete_snapshot(self, project_name, snapshot_uuid) -> Tuple[int, Dict]:
        assert project_name and snapshot_uuid, "must be provided"
        project = self.get_project(project_name)
        if not project:
            return httpstatus.NOT_FOUND, None
        proj_snaps = project.get("snapshots", None)
        if not proj_snaps:
            return httpstatus.NOT_FOUND, None
        for snap in proj_snaps:
            if snap["UUID"] == snapshot_uuid:
                proj_snaps.remove(snap)
        return httpstatus.OK, snap

    def get_snapshot_by_name(self, project_name,
                             snapshot_name) -> Tuple[int, Dict]:
        assert project_name and snapshot_name, "must be provided"
        project = self.get_project(project_name)
        if not project:
            return httpstatus.NOT_FOUND, None
        proj_snaps = project.get("snapshots", None)
        if not proj_snaps:
            return httpstatus.NOT_FOUND, None
        snap = next(iter([snap for snap in proj_snaps
                          if snap["name"] == snapshot_name]), None)
        return (httpstatus.OK, snap) if snap else (httpstatus.NOT_FOUND, None)

    def get_snapshot_by_uuid(self, project_name,
                             snapshot_uuid) -> Tuple[int, Dict]:
        assert project_name and snapshot_uuid, "must be provided"
        project = self.get_project(project_name)
        if not project:
            return httpstatus.NOT_FOUND, None
        proj_snaps = project.get("snapshots", None)
        if not proj_snaps:
            return httpstatus.NOT_FOUND, None
        snap = next(iter([snap for snap in proj_snaps
                          if snap["UUID"] == snapshot_uuid]), None)
        return (httpstatus.OK, snap) if snap else (httpstatus.NOT_FOUND, None)


class LightOSStorageVolumeDriverTest(test.TestCase):

    def setUp(self):
        """Initialize LightOS Storage Driver."""
        super(LightOSStorageVolumeDriverTest, self).setUp()

        configuration = mock.Mock(conf.Configuration)

        configuration.lightos_api_address = \
            "10.10.10.71,10.10.10.72,10.10.10.73"
        configuration.lightos_api_port = 443
        configuration.lightos_jwt = None
        configuration.lightos_snapshotname_prefix = 'openstack_'
        configuration.lightos_intermediate_snapshot_name_prefix = 'for_clone_'
        configuration.lightos_default_compression_enabled = (
            DEFAULT_COMPRESSION)
        configuration.lightos_default_num_replicas = 3
        configuration.num_volume_device_scan_tries = (
            DEVICE_SCAN_ATTEMPTS_DEFAULT)
        configuration.lightos_api_service_timeout = LIGHTOS_API_SERVICE_TIMEOUT
        configuration.driver_ssl_cert_verify = False
        # for some reason this value is not initialized by the driver parent
        # configs
        configuration.volume_name_template = 'volume-%s'
        configuration.initiator_connector = (
            "cinder.tests.unit.volume.drivers.lightos."
            "test_lightos_storage.InitiatorConnectorFactoryMocker")
        configuration.volume_backend_name = VOLUME_BACKEND_NAME
        configuration.reserved_percentage = RESERVED_PERCENTAGE

        def mocked_safe_get(config, variable_name):
            if hasattr(config, variable_name):
                return config.__getattribute__(variable_name)
            else:
                return None

        configuration.safe_get = functools.partial(mocked_safe_get,
                                                   configuration)
        self.driver = lightos.LightOSVolumeDriver(configuration=configuration)
        self.ctxt = context.get_admin_context()
        self.db: DBMock = DBMock()

        # define a default send_cmd override to return default values.
        def send_cmd_default_mock(cmd, timeout, **kwargs):
            if cmd == "get_nodes":
                return (httpstatus.OK, FAKE_LIGHTOS_CLUSTER_NODES)
            if cmd == "get_node":
                self.assertTrue(kwargs["UUID"])
                for node in FAKE_LIGHTOS_CLUSTER_NODES["nodes"]:
                    if kwargs["UUID"] == node["UUID"]:
                        return (httpstatus.OK, node)
                return (httpstatus.NOT_FOUND, node)
            elif cmd == "get_cluster_info":
                return (httpstatus.OK, FAKE_LIGHTOS_CLUSTER_INFO)
            elif cmd == "create_volume":
                project_name = kwargs["project_name"]
                volume = {
                    "project_name": project_name,
                    "name": kwargs["name"],
                    "size": kwargs["size"],
                    "n_replicas": kwargs["n_replicas"],
                    "compression": kwargs["compression"],
                    "src_snapshot_name": kwargs["src_snapshot_name"],
                    "acl": {'values': kwargs.get('acl')},
                    "state": "Available",
                }
                volume["ETag"] = get_vol_etag(volume)
                code, new_vol = self.db.create_volume(volume)
                return (code, new_vol)
            elif cmd == "delete_volume":
                return self.db.delete_volume(kwargs["project_name"],
                                             kwargs["volume_uuid"])
            elif cmd == "get_volume":
                return self.db.get_volume_by_uuid(kwargs["project_name"],
                                                  kwargs["volume_uuid"])
            elif cmd == "get_volume_by_name":
                return self.db.get_volume_by_name(kwargs["project_name"],
                                                  kwargs["volume_name"])
            elif cmd == "extend_volume":
                size = kwargs.get("size", None)
                return self.db.update_volume_by_uuid(kwargs["project_name"],
                                                     kwargs["volume_uuid"],
                                                     size=size)
            elif cmd == "create_snapshot":
                snapshot = {
                    "project_name": kwargs.get("project_name", None),
                    "name": kwargs.get("name", None),
                    "state": "Available",
                }
                return self.db.create_snapshot(snapshot)
            elif cmd == "delete_snapshot":
                return self.db.delete_snapshot(kwargs["project_name"],
                                               kwargs["snapshot_uuid"])
            elif cmd == "get_snapshot":
                return self.db.get_snapshot_by_uuid(kwargs["project_name"],
                                                    kwargs["snapshot_uuid"])
            elif cmd == "get_snapshot_by_name":
                return self.db.get_snapshot_by_name(kwargs["project_name"],
                                                    kwargs["snapshot_name"])
            elif cmd == "update_volume":
                return self.db.update_volume_by_uuid(**kwargs)

            else:
                raise RuntimeError(
                    f"'{cmd}' is not implemented. kwargs: {kwargs}")

        self.driver.cluster.send_cmd = send_cmd_default_mock

    def test_setup_should_fail_if_lightos_client_cant_auth_cluster(self):
        """Verify lightos_client fail with bad auth."""

        def side_effect(cmd, timeout):
            if cmd == "get_cluster_info":
                return (httpstatus.UNAUTHORIZED, None)
            else:
                raise RuntimeError()

        self.driver.cluster.send_cmd = side_effect
        self.assertRaises(exception.InvalidAuthKey,
                          self.driver.do_setup, None)

    def test_setup_should_succeed(self):
        """Test that lightos_client succeed."""
        self.driver.do_setup(None)

    def test_create_volume_should_succeed(self):
        """Test that lightos_client succeed."""
        self.driver.do_setup(None)

        vol_type = test_utils.create_volume_type(self.ctxt, self,
                                                 name='my_vol_type')
        volume = test_utils.create_volume(self.ctxt, size=4,
                                          volume_type_id=vol_type.id)

        self.driver.create_volume(volume)
        self.driver.delete_volume(volume)
        db.volume_destroy(self.ctxt, volume.id)

    def test_create_volume_same_volume_twice_succeed(self):
        """Test succeed to create an exiting volume."""
        self.driver.do_setup(None)

        vol_type = test_utils.create_volume_type(self.ctxt, self,
                                                 name='my_vol_type')
        volume = test_utils.create_volume(self.ctxt, size=4,
                                          volume_type_id=vol_type.id)

        self.driver.create_volume(volume)
        self.driver.create_volume(volume)
        self.driver.delete_volume(volume)
        db.volume_destroy(self.ctxt, volume.id)

    def _create_volume_in_failed_state(self, vol_state):
        """Verify scenario of created volume in failed state:

        Driver is expected to issue a deletion command and raise exception
        """
        def send_cmd_mock(cmd, **kwargs):
            if cmd == "create_volume":
                project_name = kwargs["project_name"]
                volume = {
                    "project_name": project_name,
                    "name": kwargs["name"],
                    "size": kwargs["size"],
                    "n_replicas": kwargs["n_replicas"],
                    "compression": kwargs["compression"],
                    "src_snapshot_name": kwargs["src_snapshot_name"],
                    "acl": {'values': kwargs.get('acl')},
                    "state": vol_state,
                }
                volume["ETag"] = get_vol_etag(volume)
                code, new_vol = self.db.create_volume(volume)
                return (code, new_vol)
            else:
                return cluster_send_cmd(cmd, **kwargs)

        self.driver.do_setup(None)
        cluster_send_cmd = deepcopy(self.driver.cluster.send_cmd)
        self.driver.cluster.send_cmd = send_cmd_mock
        vol_type = test_utils.create_volume_type(self.ctxt, self,
                                                 name='my_vol_type')
        volume = test_utils.create_volume(self.ctxt, size=4,
                                          volume_type_id=vol_type.id)
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.create_volume, volume)
        proj = self.db.data["projects"][lightos.LIGHTOS_DEFAULT_PROJECT_NAME]
        actual_volumes = proj["volumes"]
        self.assertEqual(0, len(actual_volumes))
        db.volume_destroy(self.ctxt, volume.id)

    def test_create_volume_in_failed_state(self):
        self._create_volume_in_failed_state("Failed")

    def test_create_volume_in_rollback_state(self):
        self._create_volume_in_failed_state("Rollback")

    def test_create_volume_in_migrating_state_succeed(self):
        """Verify scenario of created volume in migrating state:

        Driver is expected to succeed.
        """
        def send_cmd_mock(cmd, **kwargs):
            if cmd == "create_volume":
                project_name = kwargs["project_name"]
                volume = {
                    "project_name": project_name,
                    "name": kwargs["name"],
                    "size": kwargs["size"],
                    "n_replicas": kwargs["n_replicas"],
                    "compression": kwargs["compression"],
                    "src_snapshot_name": kwargs["src_snapshot_name"],
                    "acl": {'values': kwargs.get('acl')},
                    "state": "Migrating",
                }
                volume["ETag"] = get_vol_etag(volume)
                code, new_vol = self.db.create_volume(volume)
                return (code, new_vol)
            else:
                return cluster_send_cmd(cmd, **kwargs)
        self.driver.do_setup(None)
        cluster_send_cmd = deepcopy(self.driver.cluster.send_cmd)
        self.driver.cluster.send_cmd = send_cmd_mock
        vol_type = test_utils.create_volume_type(self.ctxt, self,
                                                 name='my_vol_type')
        volume = test_utils.create_volume(self.ctxt, size=4,
                                          volume_type_id=vol_type.id)
        self.driver.create_volume(volume)
        proj = self.db.data["projects"][lightos.LIGHTOS_DEFAULT_PROJECT_NAME]
        actual_volumes = proj["volumes"]
        self.assertEqual(1, len(actual_volumes))
        db.volume_destroy(self.ctxt, volume.id)

    def test_delete_volume_fail_if_not_created(self):
        """Test that lightos_client fail creating an already exists volume."""
        self.driver.do_setup(None)

        vol_type = test_utils.create_volume_type(self.ctxt, self,
                                                 name='my_vol_type')
        volume = test_utils.create_volume(self.ctxt, size=4,
                                          volume_type_id=vol_type.id)
        self.driver.delete_volume(volume)
        db.volume_destroy(self.ctxt, volume.id)

    def test_extend_volume_should_succeed(self):
        self.driver.do_setup(None)

        vol_type = test_utils.create_volume_type(self.ctxt, self,
                                                 name='my_vol_type')
        volume = test_utils.create_volume(self.ctxt, size=4,
                                          volume_type_id=vol_type.id)

        self.driver.create_volume(volume)
        self.driver.extend_volume(volume, 6)
        self.driver.delete_volume(volume)
        db.volume_destroy(self.ctxt, volume.id)

    def test_get_volume_specs_compression_True(self):
        self.driver.do_setup(None)

        vol_type = test_utils.create_volume_type(
            self.ctxt, self,
            extra_specs={'compression': 'True'},
            name='my_vol_typ1')

        vol_type2 = test_utils.create_volume_type(
            self.ctxt, self,
            extra_specs={'compression': '<is> True'},
            name='my_vol_type2')

        vol_type3 = test_utils.create_volume_type(
            self.ctxt, self,
            name='my_vol_type3')

        volume1 = test_utils.create_volume(self.ctxt, size=4,
                                           volume_type_id=vol_type.id)
        volume2 = test_utils.create_volume(self.ctxt, size=4,
                                           volume_type_id=vol_type2.id)
        volume3 = test_utils.create_volume(self.ctxt, size=4,
                                           volume_type_id=vol_type3.id)
        compression, _, _ = self.driver._get_volume_specs(volume1)
        self.assertTrue(compression == "True")
        compression, _, _ = self.driver._get_volume_specs(volume2)
        self.assertTrue(compression == "True")
        compression, _, _ = self.driver._get_volume_specs(volume3)
        self.assertTrue(compression == "False")

        db.volume_destroy(self.ctxt, volume1.id)
        db.volume_destroy(self.ctxt, volume2.id)
        db.volume_destroy(self.ctxt, volume3.id)

    def test_get_volume_specs_compression_False(self):
        self.driver.do_setup(None)
        self.driver.configuration.lightos_default_compression_enabled = True
        vol_type = test_utils.create_volume_type(
            self.ctxt, self,
            extra_specs={'compression': 'False'},
            name='my_vol_typ1')

        vol_type2 = test_utils.create_volume_type(
            self.ctxt, self,
            extra_specs={'compression': '<is> False'},
            name='my_vol_type2')

        vol_type3 = test_utils.create_volume_type(
            self.ctxt, self,
            name='my_vol_type3')
        volume1 = test_utils.create_volume(self.ctxt, size=4,
                                           volume_type_id=vol_type.id)
        volume2 = test_utils.create_volume(self.ctxt, size=4,
                                           volume_type_id=vol_type2.id)
        volume3 = test_utils.create_volume(self.ctxt, size=4,
                                           volume_type_id=vol_type3.id)
        compression, _, _ = self.driver._get_volume_specs(volume1)
        self.assertTrue(compression == "False")
        compression, _, _ = self.driver._get_volume_specs(volume2)
        self.assertTrue(compression == "False")
        compression, _, _ = self.driver._get_volume_specs(volume3)
        self.assertTrue(compression == "True")

        db.volume_destroy(self.ctxt, volume1.id)
        db.volume_destroy(self.ctxt, volume2.id)
        db.volume_destroy(self.ctxt, volume3.id)

    def test_extend_volume_should_fail_if_volume_does_not_exist(self):
        self.driver.do_setup(None)

        vol_type = test_utils.create_volume_type(self.ctxt, self,
                                                 name='my_vol_type')
        volume = test_utils.create_volume(self.ctxt, size=4,
                                          volume_type_id=vol_type.id)

        self.assertRaises(exception.VolumeNotFound,
                          self.driver.extend_volume, volume, 6)
        db.volume_destroy(self.ctxt, volume.id)

    def test_create_snapshot(self):
        self.driver.do_setup(None)

        vol_type = test_utils.create_volume_type(self.ctxt, self,
                                                 name='my_vol_type')
        volume = test_utils.create_volume(self.ctxt, size=4,
                                          volume_type_id=vol_type.id)
        snapshot = test_utils.create_snapshot(self.ctxt, volume_id=volume.id)

        self.driver.create_volume(volume)
        self.driver.create_snapshot(snapshot)
        self.driver.delete_volume(volume)
        db.volume_destroy(self.ctxt, volume.id)

    def test_delete_snapshot(self):
        self.driver.do_setup(None)

        vol_type = test_utils.create_volume_type(self.ctxt, self,
                                                 name='my_vol_type')
        volume = test_utils.create_volume(self.ctxt, size=4,
                                          volume_type_id=vol_type.id)
        snapshot = test_utils.create_snapshot(self.ctxt, volume_id=volume.id)

        self.driver.create_volume(volume)
        self.driver.create_snapshot(snapshot)
        self.driver.delete_snapshot(snapshot)
        self.driver.delete_volume(volume)
        db.volume_destroy(self.ctxt, volume.id)

    def test_create_volume_from_snapshot(self):
        self.driver.do_setup(None)
        vol_type = test_utils.create_volume_type(self.ctxt, self,
                                                 name='my_vol_type')
        volume = test_utils.create_volume(self.ctxt, size=4,
                                          volume_type_id=vol_type.id)
        snapshot = test_utils.create_snapshot(self.ctxt, volume_id=volume.id)
        self.driver.create_volume_from_snapshot(volume, snapshot)
        proj = self.db.data["projects"][lightos.LIGHTOS_DEFAULT_PROJECT_NAME]
        actual_volumes = proj["volumes"]
        self.assertEqual(1, len(actual_volumes))
        self.driver.delete_snapshot(snapshot)
        self.driver.delete_volume(volume)
        db.volume_destroy(self.ctxt, volume.id)
        db.snapshot_destroy(self.ctxt, snapshot.id)

    def test_initialize_connection(self):
        InitialConnectorMock.nqn = "hostnqn1"
        InitialConnectorMock.found_discovery_client = True
        self.driver.do_setup(None)
        vol_type = test_utils.create_volume_type(self.ctxt, self,
                                                 name='my_vol_type')
        volume = test_utils.create_volume(self.ctxt, size=4,
                                          volume_type_id=vol_type.id)
        self.driver.create_volume(volume)
        connection_props = \
            self.driver.initialize_connection(volume,
                                              get_connector_properties())
        self.assertIn('driver_volume_type', connection_props)
        self.assertEqual('lightos', connection_props['driver_volume_type'])
        self.assertEqual(FAKE_LIGHTOS_CLUSTER_INFO['subsystemNQN'],
                         connection_props['data']['subsysnqn'])
        self.assertEqual(
            self.db.data['projects']['default']['volumes'][0]['UUID'],
            connection_props['data']['uuid'])

        self.driver.delete_volume(volume)
        db.volume_destroy(self.ctxt, volume.id)

    def test_initialize_connection_mirgrating_volume(self):
        InitialConnectorMock.nqn = "hostnqn1"
        InitialConnectorMock.found_discovery_client = True

        def send_cmd_mock(cmd, **kwargs):
            if cmd == "create_volume":
                project_name = kwargs["project_name"]
                volume = {
                    "project_name": project_name,
                    "name": kwargs["name"],
                    "size": kwargs["size"],
                    "n_replicas": kwargs["n_replicas"],
                    "compression": kwargs["compression"],
                    "src_snapshot_name": kwargs["src_snapshot_name"],
                    "acl": {'values': kwargs.get('acl')},
                    "state": "Migrating",
                }
                volume["ETag"] = get_vol_etag(volume)
                code, new_vol = self.db.create_volume(volume)
                return (code, new_vol)
            else:
                return cluster_send_cmd(cmd, **kwargs)
        self.driver.do_setup(None)
        cluster_send_cmd = deepcopy(self.driver.cluster.send_cmd)
        self.driver.cluster.send_cmd = send_cmd_mock
        vol_type = test_utils.create_volume_type(self.ctxt, self,
                                                 name='my_vol_type')
        volume = test_utils.create_volume(self.ctxt, size=4,
                                          volume_type_id=vol_type.id)
        self.driver.create_volume(volume)
        connection_props = (
            self.driver.initialize_connection(volume,
                                              get_connector_properties()))
        self.assertIn('driver_volume_type', connection_props)
        self.assertEqual('lightos', connection_props['driver_volume_type'])
        self.assertEqual(FAKE_LIGHTOS_CLUSTER_INFO['subsystemNQN'],
                         connection_props['data']['subsysnqn'])
        self.assertEqual(
            self.db.data['projects']['default']['volumes'][0]['UUID'],
            connection_props['data']['uuid'])

        self.driver.delete_volume(volume)
        db.volume_destroy(self.ctxt, volume.id)

    def test_initialize_connection_ipv6(self):
        def side_effect(cmd, timeout, **kwargs):
            if cmd == "get_nodes":
                return (httpstatus.OK, FAKE_LIGHTOS_CLUSTER_NODES_IPV6)
            else:
                return cluster_send_cmd(cmd, timeout, **kwargs)
        cluster_send_cmd = deepcopy(self.driver.cluster.send_cmd)
        self.driver.cluster.send_cmd = side_effect
        InitialConnectorMock.nqn = "hostnqn1"
        InitialConnectorMock.found_discovery_client = True
        self.driver.do_setup(None)
        vol_type = test_utils.create_volume_type(self.ctxt, self,
                                                 name='my_vol_type')
        volume = test_utils.create_volume(self.ctxt, size=4,
                                          volume_type_id=vol_type.id)
        self.driver.create_volume(volume)
        connection_props = (
            self.driver.initialize_connection(volume,
                                              get_connector_properties()))
        self.assertIn('driver_volume_type', connection_props)
        self.assertEqual('lightos', connection_props['driver_volume_type'])
        self.assertEqual(FAKE_LIGHTOS_CLUSTER_INFO['subsystemNQN'],
                         connection_props['data']['subsysnqn'])
        self.assertEqual(
            self.db.data['projects']['default']['volumes'][0]['UUID'],
            connection_props['data']['uuid'])
        for connection in connection_props['data']['lightos_nodes']:
            self.assertIn(connection, IPV6_LIST)

        self.driver.delete_volume(volume)
        db.volume_destroy(self.ctxt, volume.id)

    def test_initialize_connection_no_hostnqn_should_fail(self):
        InitialConnectorMock.nqn = ""
        InitialConnectorMock.found_discovery_client = True
        self.driver.do_setup(None)
        vol_type = test_utils.create_volume_type(self.ctxt, self,
                                                 name='my_vol_type')
        volume = test_utils.create_volume(self.ctxt, size=4,
                                          volume_type_id=vol_type.id)
        self.driver.create_volume(volume)
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.initialize_connection, volume,
                          get_connector_properties())
        self.driver.delete_volume(volume)
        db.volume_destroy(self.ctxt, volume.id)

    def test_initialize_connection_no_dsc_should_fail(self):
        InitialConnectorMock.nqn = "hostnqn1"
        InitialConnectorMock.found_discovery_client = False
        self.driver.do_setup(None)
        vol_type = test_utils.create_volume_type(self.ctxt, self,
                                                 name='my_vol_type')
        volume = test_utils.create_volume(self.ctxt, size=4,
                                          volume_type_id=vol_type.id)
        self.driver.create_volume(volume)
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.initialize_connection, volume,
                          get_connector_properties())
        self.driver.delete_volume(volume)
        db.volume_destroy(self.ctxt, volume.id)

    def test_terminate_connection_with_hostnqn(self):
        InitialConnectorMock.nqn = "hostnqn1"
        InitialConnectorMock.found_discovery_client = True
        self.driver.do_setup(None)
        vol_type = test_utils.create_volume_type(self.ctxt, self,
                                                 name='my_vol_type')
        volume = test_utils.create_volume(self.ctxt, size=4,
                                          volume_type_id=vol_type.id)
        self.driver.create_volume(volume)
        self.driver.terminate_connection(volume, get_connector_properties())
        self.driver.delete_volume(volume)
        db.volume_destroy(self.ctxt, volume.id)

    def test_terminate_connection_with_empty_hostnqn_should_fail(self):
        InitialConnectorMock.nqn = ""
        InitialConnectorMock.found_discovery_client = True
        self.driver.do_setup(None)
        vol_type = test_utils.create_volume_type(self.ctxt, self,
                                                 name='my_vol_type')
        volume = test_utils.create_volume(self.ctxt, size=4,
                                          volume_type_id=vol_type.id)
        self.driver.create_volume(volume)
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.terminate_connection, volume,
                          get_connector_properties())
        self.driver.delete_volume(volume)
        db.volume_destroy(self.ctxt, volume.id)

    def test_force_terminate_connection_with_empty_hostnqn(self):
        InitialConnectorMock.nqn = ""
        InitialConnectorMock.found_discovery_client = True
        self.driver.do_setup(None)
        vol_type = test_utils.create_volume_type(self.ctxt, self,
                                                 name='my_vol_type')
        volume = test_utils.create_volume(self.ctxt, size=4,
                                          volume_type_id=vol_type.id)
        self.driver.create_volume(volume)
        self.driver.terminate_connection(volume, get_connector_properties(),
                                         force=True)
        self.driver.delete_volume(volume)
        db.volume_destroy(self.ctxt, volume.id)

    def test_check_for_setup_error(self):
        InitialConnectorMock.nqn = "hostnqn1"
        InitialConnectorMock.found_discovery_client = True
        self.driver.do_setup(None)
        self.driver.check_for_setup_error()

    def test_check_for_setup_error_no_subsysnqn_should_fail(self):
        InitialConnectorMock.nqn = "hostnqn1"
        InitialConnectorMock.found_discovery_client = True
        self.driver.do_setup(None)
        self.driver.cluster.subsystemNQN = ""
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.check_for_setup_error)

    def test_check_for_setup_error_no_hostnqn_should_fail(self):
        InitialConnectorMock.nqn = ""
        InitialConnectorMock.found_discovery_client = True
        self.driver.do_setup(None)
        self.assertRaises(exception.VolumeBackendAPIException,
                          self.driver.check_for_setup_error)

    def test_check_ip_format(self):
        InitialConnectorMock.nqn = ""
        InitialConnectorMock.found_discovery_client = True
        self.driver.do_setup(None)
        host = "1.1.1.1"
        port = 8009
        endpoint = self.driver.cluster._format_endpoint(host, port)
        self.assertEqual("1.1.1.1:8009", endpoint)
        host = "::1111"
        endpoint = self.driver.cluster._format_endpoint(host, port)
        self.assertEqual("[::1111]:8009", endpoint)

    def test_check_for_setup_error_no_dsc_should_succeed(self):
        InitialConnectorMock.nqn = "hostnqn1"
        InitialConnectorMock.found_discovery_client = False
        self.driver.do_setup(None)
        self.driver.check_for_setup_error()

    def test_create_clone(self):
        self.driver.do_setup(None)

        vol_type = test_utils.create_volume_type(self.ctxt, self,
                                                 name='my_vol_type')
        volume = test_utils.create_volume(self.ctxt, size=4,
                                          volume_type_id=vol_type.id)
        clone = test_utils.create_volume(self.ctxt, size=4,
                                         volume_type_id=vol_type.id)

        self.driver.create_volume(volume)
        self.driver.create_cloned_volume(clone, volume)
        self.driver.delete_volume(volume)
        self.driver.delete_volume(clone)

        db.volume_destroy(self.ctxt, volume.id)
        db.volume_destroy(self.ctxt, clone.id)

    def test_get_volume_stats(self):
        """Test that lightos_client succeed."""
        self.driver.do_setup(None)
        volumes_data = self.driver.get_volume_stats(refresh=False)
        assert len(volumes_data) == 0, "Expected empty config"

        volumes_data = self.driver.get_volume_stats(refresh=True)
        assert volumes_data['vendor_name'] == 'LightOS Storage', \
            "Expected 'LightOS Storage', received %s" % \
            volumes_data['vendor_name']
        assert volumes_data['volume_backend_name'] == VOLUME_BACKEND_NAME, \
            "Expected %s, received %s" % \
            (VOLUME_BACKEND_NAME, volumes_data['volume_backend_name'])
        assert volumes_data['driver_version'] == self.driver.VERSION, \
            "Expected %s, received %s" % \
            (self.driver.VERSION, volumes_data['driver_version'])
        assert volumes_data['storage_protocol'] == "lightos", \
            "Expected 'lightos', received %s" % \
            volumes_data['storage_protocol']
        assert volumes_data['reserved_percentage'] == RESERVED_PERCENTAGE, \
            "Expected %d, received %s" % \
            (RESERVED_PERCENTAGE, volumes_data['reserved_percentage'])
        assert volumes_data['QoS_support'] is False, \
            "Expected False, received %s" % volumes_data['QoS_support']
        assert volumes_data['online_extend_support'] is True, \
            "Expected True, received %s" % \
            volumes_data['online_extend_support']
        assert volumes_data['thin_provisioning_support'] is True, \
            "Expected True, received %s" % \
            volumes_data['thin_provisioning_support']
        assert volumes_data['compression'] == [True, False], \
            "Expected [True, False], received %s" % volumes_data['compression']
        assert volumes_data['multiattach'] is True, \
            "Expected True, received %s" % volumes_data['multiattach']
        assert volumes_data['free_capacity_gb'] == 'infinite', \
            "Expected 'infinite', received %s" % \
            volumes_data['free_capacity_gb']



#   Copyright 2013, Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
from http import HTTPStatus

from oslo_log import log as logging
import webob

from cinder.api import extensions
from cinder.api.openstack import wsgi
from cinder.api.schemas import snapshot_actions
from cinder.api import validation
from cinder.i18n import _
from cinder import objects
from cinder.objects import fields
from cinder.policies import snapshot_actions as policy
LOG = logging.getLogger(__name__)


class SnapshotActionsController(wsgi.Controller):
    def __init__(self, *args, **kwargs):
        super(SnapshotActionsController, self).__init__(*args, **kwargs)
        LOG.debug("SnapshotActionsController initialized")

    @wsgi.action('os-update_snapshot_status')
    @validation.schema(snapshot_actions.update_snapshot_status)
    def _update_snapshot_status(self, req, id, body):
        """Update database fields related to status of a snapshot.

           Intended for creation of snapshots, so snapshot state
           must start as 'creating' and be changed to 'available',
           'creating', or 'error'.
        """

        context = req.environ['cinder.context']
        LOG.debug("body: %s", body)

        status = body['os-update_snapshot_status']['status']

        # Allowed state transitions
        status_map = {fields.SnapshotStatus.CREATING:
                      [fields.SnapshotStatus.CREATING,
                       fields.SnapshotStatus.AVAILABLE,
                       fields.SnapshotStatus.ERROR],
                      fields.SnapshotStatus.DELETING:
                      [fields.SnapshotStatus.DELETING,
                       fields.SnapshotStatus.ERROR_DELETING]}

        current_snapshot = objects.Snapshot.get_by_id(context, id)
        context.authorize(policy.UPDATE_STATUS_POLICY,
                          target_obj=current_snapshot)

        if current_snapshot.status not in status_map:
            msg = _("Snapshot status %(cur)s not allowed for "
                    "update_snapshot_status") % {
                        'cur': current_snapshot.status}
            raise webob.exc.HTTPBadRequest(explanation=msg)

        if status not in status_map[current_snapshot.status]:
            msg = _("Provided snapshot status %(provided)s not allowed for "
                    "snapshot with status %(current)s.") % \
                {'provided': status,
                 'current': current_snapshot.status}
            raise webob.exc.HTTPBadRequest(explanation=msg)

        update_dict = {'id': id,
                       'status': status}

        progress = body['os-update_snapshot_status'].get('progress', None)
        if progress:
            update_dict.update({'progress': progress})

        LOG.info("Updating snapshot %(id)s with info %(dict)s",
                 {'id': id, 'dict': update_dict})

        current_snapshot.update(update_dict)
        current_snapshot.save()
        return webob.Response(status_int=HTTPStatus.ACCEPTED)


class Snapshot_actions(extensions.ExtensionDescriptor):
    """Enable snapshot manager actions."""

    name = "SnapshotActions"
    alias = "os-snapshot-actions"
    updated = "2013-07-16T00:00:00+00:00"

    def get_controller_extensions(self):
        controller = SnapshotActionsController()
        extension = extensions.ControllerExtension(self,
                                                   'snapshots',
                                                   controller)
        return [extension]
